/*
 *  tn7dsl.c
 *   Avalanche DSL driver
 *
 *   Zhicheng Tang, ztang@ti.com
 *   2000 (c) Texas Instruments Inc.
 *
 *  3/23/05 CPH  Change dslhal_api_setMaxBitsPerCarrier to dslhal_api_setMaxBitsPerCarrierUpstream.
 *               Add "dslctl" to read Phy Feature Control bit.
 *  5/10/05 AV   Added tn7dsl_proc_train_mode_export(). (CQ9605).
 *  5/24/05 AV   Fixing the incorrect parsing of the DSL_BIT_TMODE environment variable after the
 *               the modulation environment variable.
 *  6/02/05 AV   Changing the location of the tn7dsl_set_modulation function to overcome the 
 *               degradation in throughput. (MR-2464).
 *  6/2/05  SANJAY Added a print in case of failue of DSL Init. Rearranged the code in case of 
 *                 Failure so that cat on proc file doesnt give Seg. fault.
 *
*/
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/atmdev.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>
#include <asm/io.h>
#include <asm/mips-boards/prom.h>
#include <linux/proc_fs.h>
#include <linux/string.h>
#include <linux/ctype.h>
#include <linux/sysctl.h>
#include <linux/timer.h>
#include <linux/vmalloc.h>

#include "tn7atm.h"
#include "tn7api.h"
#include "dsl_hal_api.h"


int testflag1 = 0;
extern int  __guDbgLevel;
extern sar_stat_t sarStat;
extern unsigned int  SAR_FREQUNCY;
static int dslInSync = 0;
static int bMarginThConfig;
static int bMarginRetrainEnable;
static char EOCVendorID[8]= {0xb5, 0x00, 0x54, 0x53, 0x54, 0x43, 0x00, 0x00};
static int trellis = 1;
static unsigned int dslReg;

#define TC_SYNC          1
#define SYNC_TIME_DELAY  500000

#ifdef CONFIG_LED_MODULE
led_reg_t ledreg[2];
static int led_on;
#endif

#define DEV_DSLMOD       1
#define MAX_STR_SIZE     256
#define DSL_MOD_SIZE     256

#ifndef TRUE
#define TRUE   1
#endif

#ifndef FALSE
#define FALSE  0
#endif

#define tn7dsl_kfree_skb(x)      dev_kfree_skb(x)


//---------------------------------------------
// Begin Clear EOC definitions
//---------------------------------------------
//#define MAX_CLEAR_EOC_BUF_SIZE   600       // moved to <tn7api.h>
#define MAX_RAW_CLEAR_EOC_MSGSIZE  514       // see ITU G997.1

#define HDLC_INIT      0xFFFF
#define HDLC_FLAG      0x7E
#define HDLC_ESC       0x7D
#define HDLC_INV       0x20
#define HDLC_GENPOL    0x8408
#define HDLC_GOOD      0xF0B8

typedef struct
{
  unsigned int  len;                               // packet length
  unsigned char   data[MAX_CLEAR_EOC_BUF_SIZE];      // data field
} clearEocBufDesc_type;

typedef struct
{
  unsigned int            clearEocEnabled;   // 0: disabled; 1: enabled
  clearEocBufDesc_type   *pTxBufDesc[4];     // Tx buffer desc pointer array
  clearEocBufDesc_type   *pRxBufDesc[4];     // Rx buffer desc pointer array
  unsigned int            txRdIndex;         // host write
  unsigned int            txWrIndex;         // DSP write
  unsigned int            rxRdIndex;         // DSP write
  unsigned int            rxWrIndex;         // host write
} clearEocParm_t;

static clearEocParm_t    ClrEoc;
static clearEocParm_t   *pClrEOC = &ClrEoc;
static int   EocRxTotalPackets;
static int   EocTxTotalPackets;
static int   EocRxTotalBytes;
static int   EocTxTotalBytes;
static int   ErrEocBufFull;
static int   ErrEocBufIndex;
static int   ErrEocBufMax;
static int   ErrEocMsgOversized;
static int   ErrEocTxHdlcCRC;
static int   ErrEocRxHdlcCRC;
static int   ErrEocRxHdlcFraming;
static int   ErrEocRxPush;
//--------------------------------------------
// End of Clear EOC definitions
//--------------------------------------------

#define host2dspAddrMapping(a)    (((int)a)&~0xe0000000)


/* For the changes for trainning modes using bit fields. 
   Defaults are for backward compatibility.
*/
#define OLD_TRAINING_VAL_GDMT   3
#define OLD_TRAINING_VAL_T1413  2
#define OLD_TRAINING_VAL_MMODE  1

#define NEW_TRAINING_VAL_GDMT   2
#define NEW_TRAINING_VAL_T1413  128
#define NEW_TRAINING_VAL_MMODE  255

typedef struct 
{
    unsigned char mode_name[20];
    unsigned int mode_value;
}dsl_modes_t;

static dsl_modes_t dsl_modes[] = { 
    {"NO_MODE", 0},
    {"ADSL_G.dmt", OLD_TRAINING_VAL_GDMT},
    {"ADSL_G.lite", 4},
    {"ADSL_G.dmt.bis", 8},
    {"ADSL_G.dmt.bis_DELT", 9},
    {"ADSL_2plus", 16},
    {"ADSL_2plus_DELT", 17},
    {"ADSL_re-adsl", 32},
    {"ADSL_re-adsl_DELT", 33},
    {"ADSL_ANSI_T1.413", OLD_TRAINING_VAL_T1413},
    {"MULTI_MODE", OLD_TRAINING_VAL_MMODE},
    {"G.shdsl", 0},
    {"IDSL", 0},
    {"HDSL", 0},
    {"SDSL", 0},
    {"VDSL", 0}    
};

/* Offsets in the dsl_modes structure */
#define G_DMT_OFFSET            1
#define G_LITE_OFFSET           2
#define G_DMT_BIS_OFFSET        3
#define G_DMT_BIS_DELT_OFFSET   4
#define ADSL_2PLUS_OFFSET       5
#define ADSL_2PLUS__DELT_OFFSET 6
#define ADSL_READSL_OFFSET      7
#define ADSL_READSL_DELT_OFFSET 8
#define T1413_OFFSET 9
#define MMODE_OFFSET 10

/* Old enum values for backward compatibility */
enum
{
  NO_MODE           = 0,
  MULTI_MODE        = 1,
  T1413_MODE        = 2,
  GDMT_MODE         = 3,
  GLITE_MODE        = 4,
  ADSL2_MODE        = 8,      
  ADSL2_DELT        = 9,      
  ADSL2_PLUS        = 16,     
  ADSL2_PLUS_DELT   = 17
};

extern int CanSuppADSL2;

#ifndef ADIAG
#define DSP_FIRMWARE_PATH "/lib/modules/ar0700xx.bin"
#else
#define DSP_FIRMWARE_PATH "/var/tmp/ar0700xx_diag.bin"
#endif

/* externs */
extern struct atm_dev *mydev;
extern unsigned int oamFarLBCount[4];
extern unsigned int dslhal_support_hostDspAddressTranslate(unsigned int addr);
extern int dslhal_support_restoreTrainingInfo(PITIDSLHW_T pIhw);
extern int dslhal_support_byteSwap32(int value);
extern int dslhal_support_getCMsgsRa(PITIDSLHW_T pIhw, char *cMsgRa);
extern int dslhal_support_blockWrite(void *buffer, void *addr, size_t count);
extern int dslhal_support_blockRead(void *addr, void *buffer, size_t count);
extern int dslhal_support_writeHostMailbox(PITIDSLHW_T pIhw, int cmd, int tag, int p1, int p2);
extern void fput(struct file *file);

/* module wide declars */
static PITIDSLHW_T    pIhw;
static char    mod_req[16]={'\t'};
static volatile int bshutdown;
static char info[MAX_STR_SIZE];
static DECLARE_MUTEX_LOCKED(adsl_sem_overlay); /* Used for DSL Polling enable     */
//kthread_t overlay_thread;
/* end of module wide declars */

/* Internal Functions */
static void tn7dsl_chng_modulation(void* data);
static unsigned int tn7dsl_set_modulation(void* data, int flag);
static void tn7dsl_ctrl_fineGain(int value);
static void tn7dsl_set_fineGainValue(int value);
//static int tn7dsl_reload_overlay(void);
static int dslmod_sysctl(ctl_table *ctl, int write, struct file * filp, void *buffer, size_t *lenp);
static void tn7dsl_register_dslss_led(void);
void tn7dsl_dslmod_sysctl_register(void);
void tn7dsl_dslmod_sysctl_unregister(void);
static int tn7dsl_clear_eoc_receive(void);
/* end of internal functions */

int os_atoi(const char *pStr)
{
  int MulNeg = (*pStr == '-' ? -1 : 1);

  return (MulNeg * (simple_strtoul(pStr, (char **)NULL, 10)));
}


int os_atoh(const char *pStr)
{
  int MulNeg = (*pStr == '-' ? -1 : 1);

  return (MulNeg * (simple_strtoul(pStr, (char **)NULL, 16)));
}

unsigned long os_atoul(const char *pStr)
{
  return((unsigned long) simple_strtoul(pStr, (char **)NULL, 10));
}

#ifdef DEBUG_BUILD
void daprintf( int uDbgLevel, char * szFmt, ...)
{
    static char     buff[256];
    va_list  ap;

    if( __guDbgLevel < uDbgLevel)
        return;

    va_start( ap, szFmt);
    vsprintf((char *)buff, szFmt, ap);
    va_end(ap);
    printk("%s", buff);
}
#endif

void dprintf( int uDbgLevel, char * szFmt, ...)
{
#ifdef DEBUG_BUILD
    static char     buff[256];
    va_list  ap;

    if( __guDbgLevel < uDbgLevel)
        return;

    va_start( ap, szFmt);
    vsprintf((char *)buff, szFmt, ap);
    va_end(ap);
    printk("%s", buff);
#endif
}

int strcmp(const char *s1, const char *s2)
{

  int size = strlen(s1);

  return(strncmp(s1, s2, size));
}

int strncmp(const char *s1, const char *s2, size_t size)
{
  int i = 0;
  int max_size = (int)size;

  while((s1[i] != 0) && i < max_size)
  {
    if(s2[i] == 0)
    {
      return -1;
    }
    if(s1[i] != s2[i])
    {  
       return 1;
    }
    i++;
  }
  if(s2[i] != 0)
  {
    return 1;
  }
  
  return 0;
}

unsigned int shim_osGetCpuFrequency(void)
{
  char *cp;
  unsigned int CpuFrequency;

  dgprintf(6, "shim_osGetCpuFrequency()\n");
  cp=(char *)prom_getenv("cpufrequency");
  CpuFrequency=os_atoi(cp);
  return CpuFrequency;
}

int shim_osLoadFWImage(char *ptr)
{
  unsigned int bytesRead;
  mm_segment_t  oldfs;
  static struct file *filp;
  unsigned int imageLength=0x5ffff;


  dgprintf(4, "tn7dsl_read_dsp()\n");

  dgprintf(4,"open file %s\n", DSP_FIRMWARE_PATH);

  filp=filp_open(DSP_FIRMWARE_PATH,00,O_RDONLY);
  if(filp ==NULL)
  {
    printk("Failed: Could not open DSP binary file\n");
          return -1;
  }

  if (filp->f_dentry != NULL)
  {
    if (filp->f_dentry->d_inode != NULL)
    {
      printk("DSP binary filesize = %d bytes\n", (int)filp->f_dentry->d_inode->i_size);
      imageLength = (unsigned int)filp->f_dentry->d_inode->i_size + 0x200;
    }
  }

  if (filp->f_op->read==NULL)
          return -1;  /* File(system) doesn't allow reads */
       
  /* Disable parameter checking */
  oldfs = get_fs();
  set_fs(KERNEL_DS);

  /* Now read bytes from postion "StartPos" */
  filp->f_pos = 0;

  bytesRead = filp->f_op->read(filp,ptr,imageLength,&filp->f_pos);

  dgprintf(4,"file length = %d\n", bytesRead);

  set_fs(oldfs);

  /* Close the file */
  fput(filp);

  return bytesRead;
}


unsigned int shim_read_overlay_page(void *ptr, unsigned int secOffset, unsigned int secLength)
{
  unsigned int bytesRead;
  mm_segment_t  oldfs;
  struct file *filp;

  dgprintf(4,"shim_read_overlay_page\n");
  //dgprintf(4,"sec offset=%d, sec length =%d\n", secOffset, secLength);

  filp=filp_open(DSP_FIRMWARE_PATH,00,O_RDONLY);
  if(filp ==NULL)
  {
    printk("Failed: Could not open DSP binary file\n");
          return -1;
  }

  if (filp->f_op->read==NULL)
          return -1;  /* File(system) doesn't allow reads */

  /* Now read bytes from postion "StartPos" */

  if(filp->f_op->llseek)
    filp->f_op->llseek(filp,secOffset, 0);
  oldfs = get_fs();
  set_fs(KERNEL_DS);
  filp->f_pos = secOffset;
  bytesRead = filp->f_op->read(filp,ptr,secLength,&filp->f_pos);

  set_fs(oldfs);
   /* Close the file */
  fput(filp);
  return bytesRead;
}

int shim_osLoadDebugFWImage(char *ptr)
{
  return 0;
}
int shim_osStringCmp(const char *s1, const char *s2)
{
  return strcmp(s1, s2);
}

void *shim_osAllocateMemory(unsigned int size)
{
  return ((void *)kmalloc(size, GFP_KERNEL));
}

void *shim_osAllocateDmaMemory(unsigned int size)
{
  /*
  int order;

  order=1;
  size=size/4096;
  while(size >= 1)
  {
    order++;
    size=size/2;
  }

  return ( (void *)__get_free_pages(GFP_ATOMIC, order));
  */
  //return ((void *)kmalloc(size, GFP_ATOMIC));
  //return ((void *)kmalloc(size, GFP_KERNEL));
  void *ptr;

  ptr = kmalloc(size, GFP_ATOMIC);
  if(ptr==NULL)
  {
    printk("failed atomic\n");
    ptr = kmalloc(size, GFP_KERNEL);
    if(ptr==NULL)
    {
      printk("failed kernel\n");
      ptr = kmalloc(size, GFP_KERNEL|GFP_DMA);
    }
  }
  //printk("size=%d\n", size);
  return ptr;

}


void shim_osFreeMemory(void *ptr, unsigned int size)
{

  kfree(ptr);
}

void shim_osFreeDmaMemory(void *ptr, unsigned int size)
{
/*
  int order;

  order=1;
  size=size/4096;
  while(size >=1)
  {
    order++;
    size=size/2;
  }
  free_pages(ptr, order);
*/
  kfree(ptr);
}

void *shim_osAllocateVMemory(unsigned int size)
{

  return ((void *)vmalloc(size));
}

void shim_osFreeVMemory(void *ptr, unsigned int size)
{
  vfree(ptr);
}

void shim_osMoveMemory(char *dst, char *src, unsigned int numBytes)
{
  memcpy(dst, src, numBytes);
}

void shim_osZeroMemory(char *dst, unsigned int numBytes)
{
  memset(dst, 0, numBytes);
}

void shim_osWriteBackCache(void *addr, unsigned int size)
{
  unsigned int i,Size=(((unsigned int)addr)&0xf)+size;

  for (i=0;i<Size;i+=16,addr+=16)
  {
    __asm__(" .set mips3 ");
    __asm__(" cache  25, (%0)" : : "r" (addr));
    __asm__(" .set mips0 ");
  }
}

void shim_osInvalidateCache(void *addr, unsigned int size)
{
  unsigned int i,Size=(((unsigned int)addr)&0xf)+size;
  if(!addr)
    return;
  for (i=0;i<Size;i+=16,addr+=16)
  {
    __asm__(" .set mips3 ");
    __asm__("cache  17, (%0)" : : "r" (addr));
    __asm__(" .set mips0 ");
  }
}

void shim_osClockWait(int val)
{
  unsigned int chkvalue;
  chkvalue=val/64;

  if(chkvalue > 1000)
    {
      mdelay(chkvalue/1000);
      return;
    }
  else
    udelay(val/64);
} /* end of cwait() */

unsigned int shim_osClockTick(int val)
{
  return jiffies;
}

int flags;
spinlock_t shimLock;

void shim_osCriticalEnter(void)
{
  spin_lock_irqsave(&shimLock, flags);

}


void shim_osCriticalExit(void)
{
  spin_unlock_irqrestore(&shimLock, flags);
}

//@Added SNR per bin info per customer request. 05-14-2004
int tn7dsl_proc_snr0(char* buf, char **start, off_t offset, int count,
                 int *eof, void *data)
{

  int len = 0;
  int limit = count - 80;
  int i, j;

  if(len<=limit)
    len += sprintf(buf+len, "\nAR7 DSL Modem Rx SNR Per Bin for Bin0 :\n");
  for (i=0; i< 512/16; i++)
  {
    for(j=0;j<16;j++)
    {
      if(len <=limit)
        len += sprintf(buf+len, "%04x ", (unsigned short)pIhw->AppData.rxSnrPerBin0[i*16+j]);
    }
    if(len <=limit)
        len += sprintf(buf+len, "\n");
  }
  return len;
}

int tn7dsl_proc_snr1(char* buf, char **start, off_t offset, int count,
                 int *eof, void *data)
{

  int len = 0;
  int limit = count - 80;
  int i, j;

  if(len<=limit)
    len += sprintf(buf+len, "\nAR7 DSL Modem Rx SNR Per Bin for Bin1 :\n");
  for (i=0; i< 512/16; i++)
  {
    for(j=0;j<16;j++)
    {
      if(len <=limit)
        len += sprintf(buf+len, "%04x ", (unsigned short)pIhw->AppData.rxSnrPerBin1[i*16+j]);
    }
    if(len <=limit)
        len += sprintf(buf+len, "\n");
  }
  return len;
}

int tn7dsl_proc_snr2(char* buf, char **start, off_t offset, int count,
                 int *eof, void *data)
{

  int len = 0;
  int limit = count - 80;
  int i, j;

  if(len<=limit)
    len += sprintf(buf+len, "\nAR7 DSL Modem Rx SNR Per Bin for Bin2 :\n");
  for (i=0; i< 512/16; i++)
  {
    for(j=0;j<16;j++)
    {
      if(len <=limit)
        len += sprintf(buf+len, "%04x ", (unsigned short)pIhw->AppData.rxSnrPerBin2[i*16+j]);
    }
    if(len <=limit)
        len += sprintf(buf+len, "\n");
  }
  return len;
}

//@Added bit allocation table per customer request. 05-14-2004
int tn7dsl_proc_bit_allocation(char* buf, char **start, off_t offset, int count,
                 int *eof, void *data)
{

  int len = 0;
  int limit = count - 80;
  int i, j;

  if(len<=limit)
    len += sprintf(buf+len, "\nAR7 DSL Modem US Bit Allocation:\n");
  for (i=0; i< 128/16; i++)
  {
    for(j=0;j<16;j++)
    {
      if(len <=limit)
        len += sprintf(buf+len, "%02x ", (unsigned char)pIhw->AppData.BitAllocTblUstrm[i*16+j]);
    }
    if(len <=limit)
        len += sprintf(buf+len, "\n");
  }

  if(len<=limit)
    len += sprintf(buf+len, "\nAR7 DSL Modem DS Bit Allocation:\n");
  for (i=0; i< 512/16; i++)
  {
    for(j=0;j<16;j++)
    {
      if(len <=limit)
        len += sprintf(buf+len, "%02x ", (unsigned char)pIhw->AppData.BitAllocTblDstrm[i*16+j]);
    }
    if(len <=limit)
        len += sprintf(buf+len, "\n");
  }

  return len;
}

#ifndef NO_ACT
int tn7dsl_proc_ds_noise(char* buf, char **start, off_t offset, int count,
                 int *eof, void *data)
{

  int len = 0;
  int limit = count - 80;
  int i, j;

  if(len<=limit)
    len += sprintf(buf+len, "\nAR7 DSL Modem DS Noise:\n");
  for (i=0; i< 512/16; i++)
  {
    for(j=0;j<16;j++)
    {
      if(len <=limit)
        len += sprintf(buf+len, "%d ", (unsigned int)pIhw->AppData.dsNoise[i*16+j]);
    }
    if(len <=limit)
        len += sprintf(buf+len, "\n");
  }
  return len;
}

#endif

int tn7dsl_proc_stats(char* buf, char **start, off_t offset, int count,
                 int *eof, void *data)
{

  int len = 0;
  int limit = count - 80;
  int F4count, F5count;
  unsigned int maxRate=0;
  unsigned int us_maxRate=0;
  /* Read Ax5 Stats */

  dslhal_api_gatherStatistics(pIhw);
  if(len<=limit)
    len += sprintf(buf+len, "\nAR7 DSL Modem Statistics:\n");
  if(len<=limit)
    len += sprintf(buf+len, "--------------------------------\n");
  /* us and ds Connection Rates */
  if(len<=limit)
    len += sprintf(buf+len, "[DSL Modem Stats]\n");


  if(len<=limit)
  {
    if(pIhw->lConnected != 1)
    {
      pIhw->AppData.USConRate = 0;
      pIhw->AppData.DSConRate = 0;
    }
    len += sprintf(buf+len, "\tUS Connection Rate:\t%u\tDS Connection Rate:\t%u\n",
                   (unsigned int)pIhw->AppData.USConRate,
                   (unsigned int)pIhw->AppData.DSConRate );
  }
  if(len<=limit)
    len += sprintf(buf+len, "\tDS Line Attenuation:\t%u\tDS Margin:\t\t%u\n",
                   (unsigned int)pIhw->AppData.dsLineAttn/2,
                   (unsigned int)pIhw->AppData.dsMargin/2 );
  if(len<=limit)
    len += sprintf(buf+len, "\tUS Line Attenuation:\t%u\tUS Margin:\t\t%u\n",
                   (unsigned int)pIhw->AppData.usLineAttn/2,
                   (unsigned int)pIhw->AppData.usMargin );
  if(len<=limit)
    len += sprintf(buf+len, "\tUS Payload :\t\t%u\tDS Payload:\t\t%u\n",
                   ((unsigned int)pIhw->AppData.usAtm_count[0] + (unsigned int)pIhw->AppData.usAtm_count[1])*48,
                   ((unsigned int)pIhw->AppData.dsGood_count[0] + (unsigned int)pIhw->AppData.dsGood_count[1])*48);
  /* Superframe Count */
  if(len<=limit)
    len += sprintf(buf+len, "\tUS Superframe Cnt :\t%u\tDS Superframe Cnt:\t%u\n",
                   (unsigned int)pIhw->AppData.usSuperFrmCnt,
                   (unsigned int)pIhw->AppData.dsSuperFrmCnt );

  /* US and DS power */
  if(len<=limit)
  {
    if(pIhw->AppData.bState < 5)
    {
      pIhw->AppData.usTxPower = 0;
      pIhw->AppData.dsTxPower = 0;
    }
    len += sprintf(buf+len, "\tUS Transmit Power :\t%u\tDS Transmit Power:\t%u\n",
                   (unsigned int)pIhw->AppData.usTxPower/256,
                   (unsigned int)pIhw->AppData.dsTxPower/256 );
  }
  /* DSL Stats Errors*/
  if(len<=limit)
    len += sprintf(buf+len, "\tLOS errors:\t\t%u\tSEF errors:\t\t%u\n",
                   (unsigned int)pIhw->AppData.LOS_errors,
                   (unsigned int)pIhw->AppData.SEF_errors );
  if(len<=limit)
    len += sprintf(buf+len, "\tFrame mode:\t\t%u\tMax Frame mode:\t\t%u\n",
                   (unsigned int)pIhw->AppData.FrmMode,
                   (unsigned int)pIhw->AppData.MaxFrmMode );
  if(len<=limit)
    len += sprintf(buf+len, "\tTrained Path:\t\t%u\tUS Peak Cell Rate:\t%u\n",
                   (unsigned int)pIhw->AppData.TrainedPath,
                   (unsigned int)pIhw->AppData.USConRate*1000/8/53 );
  if(len<=limit)
    len += sprintf(buf+len, "\tTrained Mode:\t\t%u\tSelected Mode:\t\t%u\n",
                   (unsigned int)pIhw->AppData.TrainedMode, (unsigned int)pIhw->AppData.StdMode );

  if(len<=limit)
    len += sprintf(buf+len, "\tATUC Vendor Code:\t%X\tATUC Revision:\t%u\n",
                   (unsigned int)pIhw->AppData.atucVendorId, pIhw->AppData.atucRevisionNum);
  if(len<=limit)
    len += sprintf(buf+len, "\tHybrid Selected:\t%u\tTrellis:\t\t%u\n",
                   (unsigned int)pIhw->AppData.currentHybridNum, trellis);

  //@Added Maximum attainable bit rate information. 05-14-2004
  if(pIhw->AppData.dsl_modulation > 5)
  {
    memcpy(&maxRate, &pIhw->adsl2TrainingMessages.rParams[6],4);
    maxRate /= 1000;
    memcpy(&us_maxRate, &pIhw->adsl2TrainingMessages.cParams[6],4);
  }
  else
  {
     int offset[2] = {5, 1};
     unsigned char rMsgsRA[12];
     int numPayloadBytes = 0;

     dslhal_api_dspInterfaceRead(pIhw, (unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset, (unsigned char *)&rMsgsRA[0], 12);

     maxRate = (unsigned int)pIhw->AppData.DSConRate;

     numPayloadBytes  = ((rMsgsRA[6] & 0x01) << 8) | rMsgsRA[5];
     numPayloadBytes  = dslhal_support_byteSwap32((unsigned int)numPayloadBytes);
     numPayloadBytes *= 32;

     if (numPayloadBytes > maxRate)
     {
        maxRate = numPayloadBytes;
     }
  }

  if(len<=limit)
    len += sprintf(buf+len, "\tShowtime Count:\t\t%u\tDS Max Attainable Bit Rate:\t%u\n",
                   (unsigned int)pIhw->AppData.showtimeCount, maxRate);

  if(len<=limit)
  {
    int offset[2] = {32, 0};
    unsigned int usBitswap, dsBitswap;

    tn7dsl_generic_read(2, (unsigned int *)&offset);
    dsBitswap = dslReg & dslhal_support_byteSwap32(0x000000ff);

    offset[0] = 33;
    tn7dsl_generic_read(2, (unsigned int *)&offset);
    usBitswap = dslReg & dslhal_support_byteSwap32(0x000000ff);

    if(pIhw->AppData.dsl_modulation > 5)
       len += sprintf(buf+len, "\tBitSwap:\t\t%u\tUS Max Attainable Bit Rate:\t%u\n",
                   (unsigned int)(usBitswap && dsBitswap), us_maxRate);
    else
       len += sprintf(buf+len, "\tBitSwap:\t\t%u\tUS Max Attainable Bit Rate:\tn/a\n",
                   (unsigned int)(usBitswap && dsBitswap));
  }

  /* Upstream Interleaved Errors */
  if(len<=limit)
    len += sprintf(buf+len, "\n\t[Upstream (TX) Interleave path]\n");
  if(len<=limit)
    len += sprintf(buf+len, "\tCRC: \t%u\tFEC: \t%u\tNCD: \t%u\n",
                   (unsigned int)pIhw->AppData.usICRC_errors,
                   (unsigned int)pIhw->AppData.usIFEC_errors,
                   (unsigned int)pIhw->AppData.usINCD_error);
  if(len<=limit)
    len += sprintf(buf+len, "\tLCD: \t%u\tHEC: \t%u\n",
                   (unsigned int)pIhw->AppData.usILCD_errors,
                   (unsigned int)pIhw->AppData.usIHEC_errors);
  /* Downstream Interleaved Errors */
  if(len<=limit)
    len += sprintf(buf+len, "\n\t[Downstream (RX) Interleave path]\n");
  if(len<=limit)
    len += sprintf(buf+len, "\tCRC: \t%u\tFEC: \t%u\tNCD: \t%u\n",
                   (unsigned int)pIhw->AppData.dsICRC_errors,
                   (unsigned int)pIhw->AppData.dsIFEC_errors,
                   (unsigned int)pIhw->AppData.dsINCD_error);
  if(len<=limit)
    len += sprintf(buf+len, "\tLCD: \t%u\tHEC: \t%u\n",
                   (unsigned int)pIhw->AppData.dsILCD_errors,
                   (unsigned int)pIhw->AppData.dsIHEC_errors);
  /* Upstream Fast Errors */
  if(len<=limit)
    len += sprintf(buf+len, "\n\t[Upstream (TX) Fast path]\n");
  if(len<=limit)
    len += sprintf(buf+len, "\tCRC: \t%u\tFEC: \t%u\tNCD: \t%u\n",
                   (unsigned int)pIhw->AppData.usFCRC_errors,
                   (unsigned int)pIhw->AppData.usFFEC_errors,
                   (unsigned int)pIhw->AppData.usFNCD_error);
  if(len<=limit)
    len += sprintf(buf+len, "\tLCD: \t%u\tHEC: \t%u\n",
                   (unsigned int)pIhw->AppData.usFLCD_errors,
                   (unsigned int)pIhw->AppData.usFHEC_errors);
  /* Downstream Fast Errors */
  if(len<=limit)
    len += sprintf(buf+len, "\n\t[Downstream (RX) Fast path]\n");
  if(len<=limit)
    len += sprintf(buf+len, "\tCRC: \t%u\tFEC: \t%u\tNCD: \t%u\n",
                   (unsigned int)pIhw->AppData.dsFCRC_errors,
                   (unsigned int)pIhw->AppData.dsFFEC_errors,
                   (unsigned int)pIhw->AppData.dsFNCD_error);
  if(len<=limit)
    len += sprintf(buf+len, "\tLCD: \t%u\tHEC: \t%u\n",
                   (unsigned int)pIhw->AppData.dsFLCD_errors,
                   (unsigned int)pIhw->AppData.dsFHEC_errors);

  /* ATM stats upstream */
  if(len<=limit)
    len += sprintf(buf+len, "\n[ATM Stats]");
  if(len<=limit)
    len += sprintf(buf+len, "\n\t[Upstream/TX]\n");
  if(len<=limit)
    len += sprintf(buf+len, "\tGood Cell Cnt:\t%u\n\tIdle Cell Cnt:\t%u\n\n",
                   (unsigned int)pIhw->AppData.usAtm_count[0] + (unsigned int)pIhw->AppData.usAtm_count[1],
                   (unsigned int)pIhw->AppData.usIdle_count[0] + (unsigned int)pIhw->AppData.usIdle_count[1]);
  /* ATM stats downstream */
  if(len<=limit)
    len += sprintf(buf+len, "\n\t[Downstream/RX)]\n");
  if(len<=limit)
    len += sprintf(buf+len, "\tGood Cell Cnt:\t%u\n\tIdle Cell Cnt:\t%u\n\tBad Hec Cell Cnt:\t%u\n",
                   (unsigned int)pIhw->AppData.dsGood_count[0] + (unsigned int)pIhw->AppData.dsGood_count[1],
                   (unsigned int)pIhw->AppData.dsIdle_count[0] + (unsigned int)pIhw->AppData.dsIdle_count[1],
                   (unsigned int)pIhw->AppData.dsBadHec_count[0] + (unsigned int)pIhw->AppData.dsBadHec_count[1]);
  if(len<=limit)
    len += sprintf(buf+len, "\tOverflow Dropped Cell Cnt:\t%u\n",
                   (unsigned int)pIhw->AppData.dsOVFDrop_count[0] + (unsigned int)pIhw->AppData.dsOVFDrop_count[1]);
  tn7sar_get_stats(pIhw->pOsContext);
  if(len<=limit)
    len += sprintf(buf+len, "\n[SAR AAL5 Stats]\n");
  if(len<=limit)
    len += sprintf(buf+len, "\tTx PDU's:\t%u\n\tRx PDU's:\t%u\n",
                   sarStat.txPktCnt,
                   sarStat.rxPktCnt);
  if(len<=limit)
    len += sprintf(buf+len, "\tTx Total Bytes:\t%u\n\tRx Total Bytes:\t%u\n",
                   sarStat.txBytes,
                   sarStat.rxBytes);
  if(len<=limit)
    len += sprintf(buf+len, "\tTx Total Error Counts:\t%u\n\tRx Total Error Counts:\t%u\n\n",
                   sarStat.txErrors,
                   sarStat.rxErrors);

  /* oam loopback info */
  if(len<=limit)
    len += sprintf(buf+len, "\n[OAM Stats]\n");

  tn7sar_get_near_end_loopback_count(&F4count, &F5count);

  if(len<=limit)
    {
    len += sprintf(buf+len, "\tNear End F5 Loop Back Count:\t%u\n\tNear End F4 Loop Back Count:\t%u\n\tFar End F5 Loop Back Count:\t%u\n\tFar End F4 Loop Back Count:\t%u\n",
                  F5count,
                  F4count,
                  oamFarLBCount[0] + oamFarLBCount[2],
                  oamFarLBCount[1] + oamFarLBCount[3]);
    }
  return len;
}

int tn7dsl_proc_modem(char* buf, char **start, off_t offset, int count,
                 int *eof, void *data)
{

  int len = 0;
  int limit = count - 80;

  char *state;
  int tag;

  tag= dslhal_api_pollTrainingStatus(pIhw);
  tag = pIhw->AppData.bState;

  switch(tag){
    case 0:       state = "ACTREQ";     break;
    case 1:       state = "QUIET1";     break;
    case 2:       state = "IDLE";       break;
    case 3:       state = "INIT";       break;
    case 4:       state = "RTDL";       break;
    case 5:       state = "SHOWTIME";   break;
    default:      state = "unknown";    break;
  }

  if(pIhw->lConnected == 1)
    state = "SHOWTIME";
  if(len<=limit)
    len += sprintf(buf+len,"%s\n",state);
  if(len<=limit)
    len += sprintf(buf+len, "%d\n", dslReg);

  return len;
}

/**********************************************************************
**                                                                    *
**  tn7dsl_hdlc_update_crc() -- Calculate CRC                         *
**                                                                    *
**  Input Parms:                                                      *
**      new_byte  --  data byte to be counted into CRC                *
**      crc_reg   --  current CRC value before the new_byte           *
**                                                                    *
**  Output:                                                           *
**      --                                                            *
**                                                                    *
**  Return:                                                           *
**      new_crc   --  new CRC value after new_byte is calculated in   *
**                                                                    *
***********************************************************************/
static unsigned short tn7dsl_hdlc_update_crc(unsigned char new_byte, unsigned short crc_reg)
{
  int i;

  crc_reg = crc_reg ^ new_byte;
  for (i=0; i<8; i++)
     if (crc_reg & 0x0001)
        crc_reg = (crc_reg>>1) ^ HDLC_GENPOL;
     else
        crc_reg = crc_reg >> 1;

  return (crc_reg);
}

/**********************************************************************
**                                                                    *
**  tn7dsl_hdlc_tx_process()  -- HDLC Framing                         *
**                                                                    *
**  Input Parms:                                                      *
**      data   --   pointer to HDLC framed data btyes (output)        *
**      msg    --   pointer to data bytes to be HDLC framed (input)   *
**      len    --   length of data bytes pointed to by msg            *
**  Output:                                                           *
**      data   --   pointer to HDLC frame generated                   *
**      len    --   length of data bytes in the HDLC frame pointed to *
**                  by 'data' pointer                                 *
**                                                                    *
**  Return:                                                           *
**      0  --  Success                                                *
**      1  --  Failure                                                *
**                                                                    *
***********************************************************************/
static int tn7dsl_hdlc_tx_process(unsigned char *data, unsigned char *msg, unsigned int *len)
{
  unsigned short int crc; /* 16 bit unsigned integer */
  int      N, j, msglen;
  char     temp;

  if ((data==NULL) || (msg==NULL))
  {
     return(1);
  }

  msglen = *len;

  data[0] = 0x7E;
  crc     = HDLC_INIT;
  data[1] = 0xFF;
  crc     = tn7dsl_hdlc_update_crc(data[1], crc);
  data[2] = 0x03;
  crc     = tn7dsl_hdlc_update_crc(data[2], crc);

  data[3] = 0x81;      // SNMP protocol identifier -- G.997.1
  crc     = tn7dsl_hdlc_update_crc(data[3], crc);
  data[4] = 0x4C;      // SNMP protocol identifier -- G.997.1
  crc     = tn7dsl_hdlc_update_crc(data[4], crc);

  N = 5;
  j = 0;
  while (j < msglen)
  {
     temp = msg[j++];
     crc = tn7dsl_hdlc_update_crc(temp, crc);
     if ((temp == HDLC_FLAG) || (temp == HDLC_ESC))
     {
        data[N] = HDLC_ESC;
        data[N+1] = temp ^ HDLC_INV;
        N = N + 2;
     }
     else
     {
        data[N] = temp;
        N = N + 1;
     }
  }

  crc = ~crc;

  for (j=0; j<2; crc >>= 8, j++)
  {
     temp = crc & 0x00FF;
     if ((temp == HDLC_FLAG) || (temp == HDLC_ESC))
     {
        data[N] = HDLC_ESC;
        data[N+1] = temp ^ HDLC_INV;
        N = N + 2;
     }
     else
     {
        data[N] = temp;
        N = N + 1;
     }
  }

  data[N] = 0x7E;

  *len = N + 1;

  return(0);
}

/**********************************************************************
**                                                                    *
**  tn7dsl_hdlc_rx_process()  -- HDLC De-Framing                      *
**                                                                    *
**  Input Parms:                                                      *
**      data   --   pointer to HDLC-deframed data bytes (output)      *
**      msg    --   pointer to HDLC frame (input)                     *
**      len    --   length of data bytes pointed to by msg            *
**  Output:                                                           *
**      data   --   pointer to HDLC de-framed                         *
**      len    --   length of data bytes pointed to by 'data'         *
**                                                                    *
**  Return:                                                           *
**      0  --  Success                                                *
**      1  --  Failure                                                *
**                                                                    *
***********************************************************************/
static int tn7dsl_hdlc_rx_process(unsigned char *data, unsigned char *msg, unsigned int *len)
{
  unsigned short int crc; /* 16 bit unsigned integer */
  unsigned char  rxbyte;
  int      N, j, msglen;
  int      ret = 1;

  if ((data==NULL) || (msg==NULL))
  {
     return(1);
  }

  msglen = *len;

  rxbyte = msg[0];
  if (rxbyte != HDLC_FLAG)
  {
     return(ret);   // invalid HDLC frame
  }

  if ((msg[1] != 0xFF) || (msg[2] != 0x03) ||
      (msg[3] != 0x81) || (msg[4] != 0x4c))
  {
     ErrEocRxHdlcFraming++;
     return(ret);   // not an expected clear EOC frame
  }

  crc = HDLC_INIT;
  j = 1;
  N = 0;

  while (j < msglen)
  {
     rxbyte = msg[j++];
     if (rxbyte == HDLC_ESC)
     {
        rxbyte = msg[j++];
        if ((rxbyte == (HDLC_ESC ^ HDLC_INV)) ||
            (rxbyte == (HDLC_FLAG ^ HDLC_INV)))
        {
           rxbyte ^= HDLC_INV;
        }
        else
        {
           break;
        }
     }
     else if (rxbyte == HDLC_FLAG)
     {
        break;
     }

     crc = tn7dsl_hdlc_update_crc(rxbyte, crc);

     if (j > 5)
     {
       data[N] = rxbyte;
       N++;
     }
  }

  if (crc == HDLC_GOOD && N >= 2)  // since 0xFF 03 81 4C are not counted in
  {
     N   = N - 2; // uncount FCS bytes
     ret = 0;     // success
  }

  *len = N;

  return(ret);
}

int tn7dsl_proc_eoc(char* buf, char **start, off_t OffSet, int count, int *eof, void *data)
{
  int len = 0;
  int limit = count - 80;
  int offset[2] = {34, 0};  // point to buffer parameter data structure
  clearEocParm_t   peoc;

  dslhal_api_dspInterfaceRead(pIhw, (unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset, (unsigned char *)&peoc, sizeof(clearEocParm_t));

  if (len <= limit)
     len += sprintf(buf+len, "\nClear EOC Channel:\n\n");
  if (len <= limit)
     len += sprintf(buf+len, "  Enabled:\t%d\n", dslhal_support_byteSwap32(peoc.clearEocEnabled));
  if (len <= limit)
     len += sprintf(buf+len, "  TxBuf[0]:\t0x%08x\n", dslhal_support_byteSwap32((unsigned int)peoc.pTxBufDesc[0]));
  if (len <= limit)
     len += sprintf(buf+len, "  TxBuf[1]:\t0x%08x\n", dslhal_support_byteSwap32((unsigned int)peoc.pTxBufDesc[1]));
  if (len <= limit)
     len += sprintf(buf+len, "  TxBuf[2]:\t0x%08x\n", dslhal_support_byteSwap32((unsigned int)peoc.pTxBufDesc[2]));
  if (len <= limit)
     len += sprintf(buf+len, "  TxBuf[3]:\t0x%08x\n", dslhal_support_byteSwap32((unsigned int)peoc.pTxBufDesc[3]));
  if (len <= limit)
     len += sprintf(buf+len, "  RxBuf[0]:\t0x%08x\n", dslhal_support_byteSwap32((unsigned int)peoc.pRxBufDesc[0]));
  if (len <= limit)
     len += sprintf(buf+len, "  RxBuf[1]:\t0x%08x\n", dslhal_support_byteSwap32((unsigned int)peoc.pRxBufDesc[1]));
  if (len <= limit)
     len += sprintf(buf+len, "  RxBuf[2]:\t0x%08x\n", dslhal_support_byteSwap32((unsigned int)peoc.pRxBufDesc[2]));
  if (len <= limit)
     len += sprintf(buf+len, "  RxBuf[3]:\t0x%08x\n", dslhal_support_byteSwap32((unsigned int)peoc.pRxBufDesc[3]));
  if (len <= limit)
     len += sprintf(buf+len, "  txRdIndex:\t%d\n", dslhal_support_byteSwap32(peoc.txRdIndex));
  if (len <= limit)
     len += sprintf(buf+len, "  txWrIndex:\t%d\n", dslhal_support_byteSwap32(peoc.txWrIndex));
  if (len <= limit)
     len += sprintf(buf+len, "  rxRdIndex:\t%d\n", dslhal_support_byteSwap32(peoc.rxRdIndex));
  if (len <= limit)
     len += sprintf(buf+len, "  rxWrIndex:\t%d\n\n", dslhal_support_byteSwap32(peoc.rxWrIndex));
  if (len <= limit)
     len += sprintf(buf+len, "  TotalTxPkts:\t%d\n", EocTxTotalPackets);
  if (len <= limit)
     len += sprintf(buf+len, "  TotalRxPkts:\t%d\n", EocRxTotalPackets);
  if (len <= limit)
     len += sprintf(buf+len, "  TotalTxBytes:\t%d\n", EocTxTotalBytes);
  if (len <= limit)
     len += sprintf(buf+len, "  TotalRxBytes:\t%d\n\n", EocRxTotalBytes);
  if (len <= limit)
     len += sprintf(buf+len, "  ErrBufFull:\t%d\n", ErrEocBufFull);
  if (len <= limit)
     len += sprintf(buf+len, "  ErrBufIndx:\t%d\n", ErrEocBufIndex);
  if (len <= limit)
     len += sprintf(buf+len, "  ErrBufMax:\t%d\n", ErrEocBufMax);
  if (len <= limit)
     len += sprintf(buf+len, "  ErrMsgMax:\t%d\n", ErrEocMsgOversized);
  if (len <= limit)
     len += sprintf(buf+len, "  ErrTxHDLC:\t%d\n", ErrEocTxHdlcCRC);
  if (len <= limit)
     len += sprintf(buf+len, "  ErrRxHDLC:\t%d\n", ErrEocRxHdlcCRC);
  if (len <= limit)
     len += sprintf(buf+len, "  ErrRxSnmp:\t%d\n", ErrEocRxHdlcFraming);
  if (len <= limit)
     len += sprintf(buf+len, "  ErrRxPush:\t%d\n\n", ErrEocRxPush);

  return len;
}

int tn7dsl_clear_eoc_setup(void)
{
  int   i;
  int   offset[2] = {34, 0};
  char *p;
  int   eoc_ptr;
  clearEocParm_t   peoc;

  dslhal_api_dspInterfaceRead(pIhw, (unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset, (unsigned char *)pClrEOC, sizeof(clearEocParm_t));

  for (i=0; i<4; i++)
  {
    clearEocBufDesc_type  *pBuf;

    // Alloc buf and descriptor for Tx
    // p = kmalloc(sizeof(clearEocBufDesc_type), GFP_ATOMIC);
    p = shim_osAllocateDmaMemory(sizeof(clearEocBufDesc_type));
    if (p == NULL)
    {
      int j;

//    printk("tn7dsl_clear_eoc_setup: kmalloc() failed\n");

      for (j=i-1; j>=0; j--)
      {
         shim_osFreeDmaMemory(pClrEOC->pTxBufDesc[j], sizeof(clearEocBufDesc_type));
         shim_osFreeDmaMemory(pClrEOC->pRxBufDesc[j], sizeof(clearEocBufDesc_type));
      }

      return(1);
    }

    pBuf = (clearEocBufDesc_type *)p;
    pBuf->len  = 0;
    pClrEOC->pTxBufDesc[i] = pBuf;

    /* Alloc buf and descriptor for Rx */
    // p = kmalloc(sizeof(clearEocBufDesc_type), GFP_ATOMIC);
    p = shim_osAllocateDmaMemory(sizeof(clearEocBufDesc_type));
    if (p == NULL)
    {
      int j;

//    printk("tn7dsl_clear_eoc_setup: kmalloc() failed\n");

      shim_osFreeDmaMemory(pClrEOC->pTxBufDesc[i], sizeof(clearEocBufDesc_type));

      for (j=i-1; j>=0; j--)
      {
         shim_osFreeDmaMemory(pClrEOC->pTxBufDesc[j], sizeof(clearEocBufDesc_type));
         shim_osFreeDmaMemory(pClrEOC->pRxBufDesc[j], sizeof(clearEocBufDesc_type));
      }

      return(1);
    }

    pBuf = (clearEocBufDesc_type *)p;
    pBuf->len  = 0;
    pClrEOC->pRxBufDesc[i] = pBuf;
  }

  pClrEOC->txRdIndex = pClrEOC->txWrIndex = 0;
  pClrEOC->rxRdIndex = pClrEOC->rxWrIndex = 0;

  pClrEOC->clearEocEnabled = 1;

  for (i=0; i<4; i++)
  {
    shim_osInvalidateCache(pClrEOC->pTxBufDesc[i], sizeof(clearEocBufDesc_type));

    eoc_ptr = host2dspAddrMapping((unsigned int)pClrEOC->pTxBufDesc[i]);
    peoc.pTxBufDesc[i] = (clearEocBufDesc_type *)dslhal_support_byteSwap32(eoc_ptr);
//  printk("pTxBufDesc[%d] = 0x%08x\n", i, (unsigned int)peoc.pTxBufDesc[i]);
  }

  for (i=0; i<4; i++)
  {
    shim_osInvalidateCache(pClrEOC->pRxBufDesc[i], sizeof(clearEocBufDesc_type));

    eoc_ptr = host2dspAddrMapping((unsigned int)pClrEOC->pRxBufDesc[i]);
    peoc.pRxBufDesc[i] = (clearEocBufDesc_type *)dslhal_support_byteSwap32(eoc_ptr);
//  printk("pRxBufDesc[%d] = 0x%08x\n", i, (unsigned int)peoc.pRxBufDesc[i]);
  }

  peoc.clearEocEnabled = dslhal_support_byteSwap32(1);
//printk("clearEocEnabled = %d\n", peoc.clearEocEnabled);

  peoc.txRdIndex = peoc.txWrIndex = 0;
  peoc.rxRdIndex = peoc.rxWrIndex = 0;

  // write the clear eoc memory ptr to DSP
  dslhal_api_dspInterfaceWrite(pIhw,(unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset, (unsigned char *)&peoc, sizeof(clearEocParm_t));

  return(0);
}

int tn7dsl_clear_eoc_close(void)
{
  int    i;
  int    offset[2] = {34, 0};
  clearEocBufDesc_type  *p;

  EocRxTotalPackets  = 0;
  EocTxTotalPackets  = 0;
  EocRxTotalBytes    = 0;
  EocTxTotalBytes    = 0;

  ErrEocBufFull      = 0;
  ErrEocBufIndex     = 0;
  ErrEocBufMax       = 0;
  ErrEocMsgOversized = 0;
  ErrEocTxHdlcCRC    = 0;
  ErrEocRxHdlcCRC    = 0;
  ErrEocRxHdlcFraming= 0;
  ErrEocRxPush       = 0;

  for (i=0; i<4; i++)
  {
     p = pClrEOC->pTxBufDesc[i];
     if (p)
     {
        shim_osFreeDmaMemory((char *)p, sizeof(clearEocBufDesc_type));
     }

     pClrEOC->pTxBufDesc[i] = 0;

     p = pClrEOC->pRxBufDesc[i];
     if (p)
     {
        shim_osFreeDmaMemory((char *)p, sizeof(clearEocBufDesc_type));
     }

     pClrEOC->pRxBufDesc[i] = 0;
  }

  pClrEOC->clearEocEnabled = 0;
  pClrEOC->txRdIndex = pClrEOC->txWrIndex = 0;
  pClrEOC->rxRdIndex = pClrEOC->rxWrIndex = 0;

  dslhal_api_dspInterfaceWrite(pIhw,(unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset, (unsigned char *)pClrEOC, sizeof(clearEocParm_t));

  return(0);
}

int tn7dsl_clear_eoc_send(unsigned char *data, unsigned int len)
{
  unsigned int   indx;
  unsigned int   index[2];
  unsigned int   enabled;
  int            offset0[2] = {34, 9};  // Tx buffer Read Index
  int            offset[2]  = {34, 10}; // Tx buffer Write Index
  int            offset2[2] = {34, 0};  // point to buffer parameter data structure
  int            ret = 0;

  if (len > MAX_RAW_CLEAR_EOC_MSGSIZE)
  {
     ErrEocMsgOversized++;
//   printk("tn7dsl_clear_eoc_send: ClearEOC max packet size is exceeded\n");
     return(1);
  }

  dslhal_api_dspInterfaceRead(pIhw,(unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset2, (unsigned char *)&enabled, 4);
  if (!enabled)
  {
//   printk("tn7dsl_clear_eoc_send: ClearEOC not enabled\n");
     return(1);
  }

  dslhal_api_dspInterfaceRead(pIhw,(unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset0, (unsigned char *)&index[0], 8);
  pClrEOC->txRdIndex = dslhal_support_byteSwap32(index[0]);
  indx = dslhal_support_byteSwap32(index[1]);

  if (indx >= 4)
  {
     ErrEocBufIndex++;
//   printk("tn7dsl_clear_eoc_send: Invalid txWrIndex (=%d)\n", indx);
     return(1);
  }

  if (((indx+1)%4) == pClrEOC->txRdIndex)
  {
     ErrEocBufFull++;
//   printk("tndsl_clear_eoc_send: buffer full\n");
     return(1);
  }

  // HDLC framing, calculate FCS and pad HDLC
  ret = tn7dsl_hdlc_tx_process(pClrEOC->pTxBufDesc[indx]->data, data, &len);

  if (!ret)
  {
     unsigned int wrIndex;

     pClrEOC->pTxBufDesc[indx]->len = dslhal_support_byteSwap32(len);  // set HDLC frame length

     shim_osWriteBackCache(pClrEOC->pTxBufDesc[indx], sizeof(clearEocBufDesc_type));

     pClrEOC->txWrIndex = (indx+1)%4;

     wrIndex = dslhal_support_byteSwap32(pClrEOC->txWrIndex);
     dslhal_api_dspInterfaceWrite(pIhw,(unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset, (unsigned char *)&wrIndex, 4);

     EocTxTotalPackets++;
     EocTxTotalBytes += len;
  }
  else
  {
     ErrEocTxHdlcCRC++;
   //printk("tn7dsl_clear_eoc_send: tx HDLC processing error\n");
  }

  return(ret);
}

static int tn7dsl_clear_eoc_receive(void)
{
  int               rc, ch;
  int               ret = 0;
  int               enabled;
  int               offset[2]  = {34, 11};  // Rx buffer read index
  int               offset2[2] = {34, 0};   // point to buffer parameter data structure
  void              *recvinfo, *skb;
  tn7atm_private_t *priv;
  unsigned char    *data;
  unsigned int      indx;
  unsigned int      index[2], len0, len;

  rc = 0;

  while (1)
  {
    dslhal_api_dspInterfaceRead(pIhw, (unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset, (unsigned char *)&index[0], 8);
    indx = dslhal_support_byteSwap32(index[0]);
    pClrEOC->rxWrIndex = dslhal_support_byteSwap32(index[1]);

    if (indx == pClrEOC->rxWrIndex)
    {
       return(0);
    }

    if (indx >= 4)
    {
       ErrEocBufIndex++;
     //printk("tn7dsl_clear_eoc_receive: Invalid rxRdIndex (=%d)\n", indx);
       indx = (indx + 1)%4;
       rc = 1;
       goto clrEoc_rx_exit;
    }

    dslhal_api_dspInterfaceRead(pIhw,(unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset2, (unsigned char *)&enabled, 4);
    if (!enabled)
    {
     //printk("tn7dsl_clear_eoc_receive: ClearEOC not enabled\n");
       rc = 1;
       goto clrEoc_rx_exit;
    }

    len = dslhal_support_byteSwap32(pClrEOC->pRxBufDesc[indx]->len);
    len0 = len;

    if (len > MAX_CLEAR_EOC_BUF_SIZE)
    {
     //printk("tn7dsl_clear_eoc_receive: buf size is exceeded\n");
       ErrEocBufMax++;
       rc = 1;
       goto clrEoc_rx_exit;
    }

    // chanid for clear eoc
    ch = EOC_DMA_CHAN;

    // allocate packet data memory
    priv = (tn7atm_private_t *)mydev->dev_data;
    data = (unsigned char *)tn7atm_allocate_rx_skb(mydev, &recvinfo, len, (void *)priv->lut[ch].vcc);
    if (data == NULL)
    {
     //printk("tn7dsl_clear_eoc_receive: kmalloc() failed\n");
       rc = 1;
       goto clrEoc_rx_exit;
    }

    skb = recvinfo;

    ret = tn7dsl_hdlc_rx_process(data, pClrEOC->pRxBufDesc[indx]->data, &len);

    if (ret)
    {
       //printk("tn7dsl_clear_eoc_receive: rx processing error\n");

       // free skb
       tn7atm_free_rx_skb(skb, NULL);

       ErrEocRxHdlcCRC++;
       rc = 1;
       goto clrEoc_rx_exit;
    }

    if (len > MAX_RAW_CLEAR_EOC_MSGSIZE)
    {
       ErrEocMsgOversized++;
       printk("WARNING: tn7dsl -- max_raw_clear_eoc_msgsize is exceeded\n");
    }

    // send packet to network stacks
    ret = tn7atm_receive(mydev, ch, len, skb, data);

    if (ret)
    {
     //printk("tn7dsl_clear_eoc_receive: failed to push up packet\n");
       tn7atm_free_rx_skb(skb, NULL);

       ErrEocRxPush++;
       rc = 1;
    }

    EocRxTotalPackets++;
    EocRxTotalBytes += len0;

clrEoc_rx_exit:
    // update rxRdIndex anyway
    pClrEOC->rxRdIndex = (indx+1)%4;

    pClrEOC->pRxBufDesc[indx]->len = 0;

    shim_osInvalidateCache(pClrEOC->pRxBufDesc[indx], sizeof(clearEocBufDesc_type));

    indx = dslhal_support_byteSwap32(pClrEOC->rxRdIndex);
    dslhal_api_dspInterfaceWrite(pIhw,(unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset, (unsigned char *)&indx, 4);
  } // while

  return(rc);
}

inline int tn7dsl_handle_interrupt(void)
{
  int intsrc;
  unsigned char cMsgRa[6];
  short margin;

  dgprintf(4, "tn7dsl_handle_dsl_interrupt()\n");
  if(pIhw)
  {
    intsrc=dslhal_api_acknowledgeInterrupt(pIhw);
    dslhal_api_handleTrainingInterrupt(pIhw, intsrc);

    if(pIhw->lConnected == TC_SYNC)
    {

      if(dslInSync == 0)
      {
        printk("DSL in Sync\n");
        tn7atm_device_connect_status(pIhw->pOsContext, 1);
        dslhal_api_initStatistics(pIhw);
        dslhal_api_gatherStatistics(pIhw);
#ifdef CONFIG_LED_MODULE
        led_operation(MOD_ADSL, DEF_ADSL_SYNC);
        led_on = DEF_ADSL_SYNC;
#endif
        /* add auto margin retrain */
        if(pIhw->AppData.TrainedMode < 5)
        {
          if(bMarginRetrainEnable && bMarginThConfig == 0)
          {
            dslhal_support_getCMsgsRa(pIhw, cMsgRa);
            margin = *(unsigned short *)&cMsgRa[4];
            margin = (margin >> 6) & 0x3f;
            if(margin & 0x20) // highest bit is 1
            {
              margin = -(margin & 0x1f);
            }

            //printk("margin = %d, cmsg-ra = %02x %02x %02x %02x %02x %02x\n", margin, cMsgRa[0],cMsgRa[1],cMsgRa[2],cMsgRa[3],cMsgRa[4],cMsgRa[5]);
            dslhal_api_setMarginThreshold(pIhw, margin*2); /* DSL margin is in 0.5db */
          }
        }

      }
      dslInSync = 1;
    }
    else
    {
      if(dslInSync == 1)
      {
        dslInSync = 0;
        tn7atm_device_connect_status(pIhw->pOsContext, 0);
        up(&adsl_sem_overlay);
        printk("DSL out of sync\n");
      }
#ifdef CONFIG_LED_MODULE
      if(pIhw->AppData.bState < RSTATE_INIT)
      {
        if(led_on != DEF_ADSL_IDLE)
        {
          led_operation(MOD_ADSL, DEF_ADSL_IDLE);
          led_on = DEF_ADSL_IDLE;
        }
      }
      else
      {
        if(led_on != DEF_ADSL_TRAINING)
        {
          led_operation(MOD_ADSL, DEF_ADSL_TRAINING);
          led_on = DEF_ADSL_TRAINING;
        }

      }

#endif

    }

    if (pIhw->AppData.clear_eoc)
    {
      tn7dsl_clear_eoc_receive();
      pIhw->AppData.clear_eoc = 0;
    }
  }
  return 0;
}


int tn7dsl_get_dslhal_version(char *pVer)
{
  dslVer ver;

  dslhal_api_getDslHalVersion(&ver);

  memcpy(pVer,&ver,8);
  return 0;
}

int tn7dsl_get_dsp_version(char *pVer)
{
  dspVer ver;
  dslhal_api_getDspVersion(pIhw, &ver);
  memcpy(pVer, &ver, 9);
  return 0;
}



static int tn7dsl_set_dsl(void)
{

  char  *ptr = NULL;
  int value;
  int i, offset[2]={4,11},oamFeature=0;
  char tmp[4];
  char dspVer[10];

  // OAM Feature Configuration
  dslhal_api_dspInterfaceRead(pIhw,(unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset, (unsigned char *)&oamFeature, 4);
  oamFeature |= dslhal_support_byteSwap32(0x0000000C);
  dslhal_api_dspInterfaceWrite(pIhw,(unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset, (unsigned char *)&oamFeature, 4);

  // modulation
  ptr = prom_getenv("modulation");
  if (ptr)
  {
    tn7dsl_set_modulation(ptr, FALSE);
  }

  // Fine Gains
  ptr = prom_getenv("fine_gain_control");
  if (ptr)
  {
    value = os_atoi(ptr);
    tn7dsl_ctrl_fineGain(value);
  }
  ptr = NULL;
  ptr = prom_getenv("fine_gain_value");
  if(ptr)
    tn7dsl_set_fineGainValue(os_atoh(ptr));

  // margin retrain
  ptr = NULL;
  ptr = prom_getenv("enable_margin_retrain");
  if(ptr)
  {
    value = os_atoi(ptr);
    if(value == 1)
    {
      dslhal_api_setMarginMonitorFlags(pIhw, 0, 1);
      bMarginRetrainEnable = 1;
      //printk("enable showtime margin monitor.\n");
      ptr = NULL;
      ptr = prom_getenv("margin_threshold");
      if(ptr)
      {
        value = os_atoi(ptr);
        //printk("Set margin threshold to %d x 0.5 db\n",value);
        if(value >= 0)
        {
          dslhal_api_setMarginThreshold(pIhw, value);
          bMarginThConfig=1;
        }
      }
    }
  }

  // rate adapt
  ptr = NULL;
  ptr = prom_getenv("enable_rate_adapt");
  if(ptr)
  {
    dslhal_api_setRateAdaptFlag(pIhw, os_atoi(ptr));
  }

  // trellis
  ptr = NULL;
  ptr = prom_getenv("trellis");
  if(ptr)
  {
    dslhal_api_setTrellisFlag(pIhw, os_atoi(ptr));
    trellis = os_atoi(ptr);
    //printk("trellis=%d\n");
  }

  // bitswap
  ptr = NULL;
  ptr = prom_getenv("bitswap");
  if(ptr)
  {
    int offset[2] = {33, 0};
    unsigned int bitswap;

    bitswap = os_atoi(ptr);

    tn7dsl_generic_read(2, offset);
    dslReg &= dslhal_support_byteSwap32(0xFFFFFF00);
    dslReg |= dslhal_support_byteSwap32(bitswap & 0x000000FF);
    tn7dsl_generic_write(2, offset, dslReg);

    offset[0] = 32;       // olrDspRx_p
    tn7dsl_generic_read(2, offset);
    dslReg &= dslhal_support_byteSwap32(0xFFFFFF00);
    dslReg |= dslhal_support_byteSwap32(bitswap & 0x000000FF);
    tn7dsl_generic_write(2, offset, dslReg);
    //printk("bitswap=%d\n", bitswap);
  }

  // maximum bits per carrier
  ptr = NULL;
  ptr = prom_getenv("maximum_bits_per_carrier");
  if(ptr)
  {
    dslhal_api_setMaxBitsPerCarrierUpstream(pIhw, os_atoi(ptr));
  }

  // maximum interleave depth
  ptr = NULL;
  ptr = prom_getenv("maximum_interleave_depth");
  if(ptr)
  {
    dslhal_api_setMaxInterleaverDepth(pIhw, os_atoi(ptr));
  }

  // inner and outer pairs
  ptr = NULL;
  ptr = prom_getenv("pair_selection");
  if(ptr)
  {
    dslhal_api_selectInnerOuterPair(pIhw, os_atoi(ptr));
  }

  ptr = NULL;
  ptr = prom_getenv("dgas_polarity");
  if(ptr)
  {
    dslhal_api_configureDgaspLpr(pIhw, 1, 1);
    dslhal_api_configureDgaspLpr(pIhw, 0, os_atoi(ptr));
  }

  ptr = NULL;
  ptr = prom_getenv("los_alarm");
  if(ptr)
  {
    dslhal_api_disableLosAlarm(pIhw, os_atoi(ptr));
  }

  ptr = NULL;
  ptr = prom_getenv("eoc_vendor_id");
  if(ptr)
  {
    for(i=0;i<8;i++)
    {
      tmp[0]=ptr[i*2];
      tmp[1]=ptr[i*2+1];
      tmp[2]=0;
      EOCVendorID[i] = os_atoh(tmp);
      //printk("tmp=%s--", tmp);
      //printk("ID[%d]=0x%02x ", i, (unsigned char)EOCVendorID[i]);
    }
    tn7dsl_get_dsp_version(dspVer);
    //printk("Annex =%d\n", dspVer[8]);
    if(dspVer[8]==2) // annex b
    {
      //printk("EOCVendorID=%02x %02x %02x %02x %02x %02x %02x %02x\n", EOCVendorID[0], EOCVendorID[1], EOCVendorID[2], EOCVendorID[3],
      //                                                                EOCVendorID[4], EOCVendorID[5], EOCVendorID[6], EOCVendorID[7]);
      dslhal_api_setEocVendorId(pIhw, EOCVendorID);
    }

  }
  ptr = NULL;
  ptr = prom_getenv("eoc_vendor_revision");
  if(ptr)
  {
    value = os_atoi(ptr);
    //printk("eoc rev=%d\n", os_atoi(ptr));
    dslhal_api_setEocRevisionNumber(pIhw, (char *)&value);

  }
  ptr = NULL;
  ptr = prom_getenv("eoc_vendor_serialnum");
  if(ptr)
  {
    dslhal_api_setEocSerialNumber(pIhw, ptr);
  }
  return 0;
}





/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 *  Function: static void tn7dsl_init(void)
 *
 *  Description: This function initializes
 *               Ar7 DSL interface
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

int tn7dsl_init(void *priv)
{
  char *cp = NULL;
  int retVal = 0;

  /* start dsl */
  if((retVal  = dslhal_api_dslStartup(&pIhw)) !=0 )
  {
    printk("DSL start failed. Error Code (%d)\n", retVal);
    return -1;
  }

  // set dsl into overlay page reload mode
  pIhw->bAutoRetrain = 1;

  /* For the changes for training modes using bit fields. 
     Defaults are for backward compatibility.
   */
  cp = prom_getenv("DSL_BIT_TMODE");
  if (cp)
  {
    printk("%s : env var DSL_BIT_TMODE is set\n", __FUNCTION__);
    /* indicate to the data pump to use the bitfield mode. */
    pIhw->AppData.useBitField = TRUE;
    
    dsl_modes[G_DMT_OFFSET].mode_value = NEW_TRAINING_VAL_GDMT;
    dsl_modes[T1413_OFFSET].mode_value = NEW_TRAINING_VAL_T1413;
    dsl_modes[MMODE_OFFSET].mode_value = NEW_TRAINING_VAL_MMODE;
  }
  
  // set default training properties
  tn7dsl_set_dsl();

  pIhw->pOsContext = priv;

  //start_kthread(tn7dsl_reload_overlay, &overlay_thread);

  /*register dslss LED with led module */
#ifdef CONFIG_LED_MODULE
  tn7dsl_register_dslss_led();
#endif

  pIhw->AppData.usTxPower=0;
  pIhw->AppData.dsTxPower=0;

  /* Change value, since we don't support these modes on this chip */
  if(!CanSuppADSL2)
  {
    dsl_modes[ADSL_2PLUS_OFFSET].mode_value       =  0;      
    dsl_modes[ADSL_2PLUS__DELT_OFFSET].mode_value =  0;
    dsl_modes[ADSL_READSL_OFFSET].mode_value      =  0;    
    dsl_modes[ADSL_READSL_DELT_OFFSET].mode_value =  0;
  }

  return 0;  
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 *  Function: int avsar_exit(void)
 *
 *  Description: Avalanche SAR exit function
 *
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

void tn7dsl_exit (void)
{

  bshutdown = 1;
#ifdef CONFIG_LED_MODULE
#ifdef DEREGISTER_LED
  //down(&adsl_sem_overlay);
  deregister_led_drv(LED_NUM_1);
  deregister_led_drv(LED_NUM_2);
#else
  led_operation(MOD_ADSL,DEF_ADSL_IDLE);
#endif
#endif
  //stop_kthread(&overlay_thread);
  tn7dsl_clear_eoc_close();
  dslhal_api_dslShutdown(pIhw);

}

static int tn7dsl_process_txflush_string(int *pqueue)
{
  int i=1;
  int j=0;
  int vci, vpi;
  char tmp[16];
  int chan;
  int tt;

  while(j<8)
  {
    tmp[j] = mod_req[i];
    //printk("tmp[%d]=%c, %d\n", j, tmp[j], tmp[j]);
    if(tmp[j] == 0x50 || tmp[j] == 0x70) // "p" or "P"
      break;
    j++;
    i++;
  }

  tmp[j] = 0;
  vpi = os_atoi(tmp);

  i++;
  j=0;
  while(j<8)
  {
    tmp[j] = mod_req[i];
    //printk("tmp[%d]=%c, %d\n", j, tmp[j], tmp[j]);
    if(tmp[j] == 0x43 || tmp[j] == 0x63) // "c" or "C"
      break;

    j++;
    i++;
  }

  vci = os_atoi(tmp);


  i++;
  j=0;
  tmp[j] = mod_req[i];

  while(j<8)
  {
    tmp[j] = mod_req[i];

    //printk("tmp[%d]=%c, %d\n", j, tmp[j], tmp[j]);
    if(tmp[j] == 0x51 || tmp[j] == 0x71) // "q", "Q"
      break;

    j++;
    i++;
  }
  tt = os_atoi(tmp);


  chan = tn7atm_lut_find(vpi, vci);

  *pqueue =tt;
  //printk("vci=%d, vpi=%d, queu=%d\n", vci, vpi, tt);

  return chan;
}

static int tn7dsl_process_oam_string(int *type, int *pvpi, int *pvci, int *pdelay)
{
  int i=1;
  int j=0;
  int vci, vpi;
  char tmp[16];
  int chan;
  int tt;

  while(j<8)
  {
    tmp[j] = mod_req[i];
    //printk("tmp[%d]=%c, %d\n", j, tmp[j], tmp[j]);
    if(tmp[j] == 0x50 || tmp[j] == 0x70)
      break;
    j++;
    i++;
  }

  tmp[j] = 0;
  vpi = os_atoi(tmp);

  i++;
  j=0;
  while(j<8)
  {
    tmp[j] = mod_req[i];
    //printk("tmp[%d]=%c, %d\n", j, tmp[j], tmp[j]);
    if(tmp[j] == 0x43 || tmp[j] == 0x63)
      break;

    j++;
    i++;
  }

  vci = os_atoi(tmp);

  if(vci==0) // f4 oam
    *type = 1;
  else
    *type = 0;


  tt=5000;
  i++;
  j=0;
  tmp[j] = mod_req[i];
  if(tmp[j]==0x44 || tmp[j]==0x64)
  {
    i++;
    while(j<8)
    {
      tmp[j] = mod_req[i];

      //printk("tmp[%d]=%c, %d\n", j, tmp[j], tmp[j]);
      if(tmp[j] == 0x54 || tmp[j] == 0x74)
        break;

      j++;
      i++;
    }
    tt = os_atoi(tmp);
  }

  chan = tn7atm_lut_find(vpi, vci);

  *pvci=vci;
  *pvpi=vpi;
  *pdelay =tt;
  dgprintf(2, "oam chan=%d, type =%d\n", chan, *type);

  return chan;
}

static void tn7dsl_dump_memory(void)
{
  unsigned int *pUi;
  int i=1;
  int j=0;
  int addr, len;
  char tmp[16];


  while(j<8)
  {
    tmp[j] = mod_req[i];
    j++;
    i++;
  }

  tmp[j] = 0;

  addr = os_atoh(tmp);

  printk("start dump address =0x%x\n", addr);
  pUi = (unsigned int *)addr;
  i++;
  j=0;
  while(j<8)
  {
    tmp[j] = mod_req[i];
    //printk("tmp[%d]=%c, %d\n", j, tmp[j], tmp[j]);
    if(tmp[j] == 0x43 || tmp[j] == 0x63)
      break;

    j++;
    i++;
  }

  len = os_atoi(tmp);
  j=0;
  for(i=0; i<len; i++)
  {
    if(j==0)
      printk("0x%08x: ", (unsigned int)pUi);
    printk("%08x ", *pUi);
    pUi++;
    j++;
    if(j==4)
    {
      printk("\n");
      j=0;
    }
  }

}



static int dslmod_sysctl(ctl_table *ctl, int write, struct file * filp,
      void *buffer, size_t *lenp)
{
  char *ptr;
  int ret, len = 0;
  int chan;
  int type;
  int vpi,vci,timeout,queue;
  int i;
  phySettings curr_setting;

  if (!*lenp || (filp->f_pos && !write)) {
    *lenp = 0;
    return 0;
    }
  /* DSL MODULATION is changed */
  if(write)
    {
    ret = proc_dostring(ctl, write, filp, buffer, lenp);

    switch (ctl->ctl_name)
      {
      case DEV_DSLMOD:
      ptr = strpbrk(info, " \t");
      strcpy(mod_req, info);

      /* parse the string to determine the action */
      if(mod_req[0] == 0x45 || mod_req[0] == 0x65 ) // 'e', or 'E' f5 end to end
      {
        chan = tn7dsl_process_oam_string(&type, &vpi, &vci, &timeout);
        tn7sar_oam_generation(pIhw->pOsContext, chan, type, vpi, vci, timeout);
      }
      else if(mod_req[0] == 0x53 || mod_req[0] == 0x73 ) // 's', or 'S' f5 seg to seg
      {
        chan=tn7dsl_process_oam_string(&type, &vpi, &vci,&timeout);
        type = type | (1<<1);
        tn7sar_oam_generation(pIhw->pOsContext, chan, type, vpi, vci,timeout);
      }
      else if(mod_req[0] == 0x46 || mod_req[0] == 0x66 ) // 'f', or 'F' f5 seg to seg
      {
        chan=tn7dsl_process_txflush_string(&queue);
        //printk("chan=%d, queue=%d\n", chan, queue);
        if(chan < 16)
          tn7sar_tx_flush(pIhw->pOsContext, chan, queue, 0);
      }
      else if(!strcmp(info, "dspfreq"))
      {
        printk("dsp_freq = %d\n", SAR_FREQUNCY * 4);
      }
      else if (!strcmp(info,"dslctl"))
      {
        for (i=0; i<2; i++)
        {
          memset(&curr_setting, 0, sizeof(curr_setting));
          dslhal_api_readPhyFeatureSettings(pIhw,
                            INTEROP_FEATURELIST0_PARAMID+i,
                           (void *) &curr_setting);
          printk("phy_feature_%d=0x%08x, phy_cntl_mask_%d=0x%08x\n",
            i, curr_setting.phyEnableDisableWord,
            i, curr_setting.phyControlWord);
        }
      }

    //debug only. Dump memory
      else if(mod_req[0] == 0x44 || mod_req[0] == 0x64 ) // 'd' or 'D'
        tn7dsl_dump_memory();
      else
        tn7dsl_chng_modulation(info);
      break;
      }
    }
  else
    {
    len += sprintf(info+len, mod_req);
    ret = proc_dostring(ctl, write, filp, buffer, lenp);
    }
  return ret;
}


ctl_table dslmod_table[] = {
  {DEV_DSLMOD, "dslmod", info, DSL_MOD_SIZE, 0644, NULL, &dslmod_sysctl},
  {0}
  };

/* Make sure that /proc/sys/dev is there */
ctl_table dslmod_root_table[] = {
#ifdef CONFIG_PROC_FS
  {CTL_DEV, "dev", NULL, 0, 0555, dslmod_table},
#endif /* CONFIG_PROC_FS */
  {0}
  };

static struct ctl_table_header *dslmod_sysctl_header;

void tn7dsl_dslmod_sysctl_register(void)
{
  static int initialized;

  if (initialized == 1)
    return;

  dslmod_sysctl_header = register_sysctl_table(dslmod_root_table, 1);
  dslmod_root_table->child->de->owner = THIS_MODULE;

  /* set the defaults */
  info[0] = 0;

  initialized = 1;
}

void tn7dsl_dslmod_sysctl_unregister(void)
{
  unregister_sysctl_table(dslmod_sysctl_header);
}



/* Codes added for compiling tiadiag.o for Analog Diagnostic tests */
#ifdef ADIAG

enum
{
  HOST_ACTREQ,         // Send R-ACKREQ and monitor for C-ACKx
  HOST_QUIET,          // Sit quietly doing nothing for about 60 seconds, DEFAULT STATE; R_IDLE
  HOST_XMITBITSWAP,    // Perform upstream bitswap - FOR INTERNAL USE ONLY
  HOST_RCVBITSWAP,     // Perform downstream bitswap - FOR INTERNAL USE ONLY
  HOST_RTDLPKT,        // Send a remote download packet - FOR INTERNAL USE ONLY
  HOST_CHANGELED,      // Read the LED settings and change accordingly
  HOST_IDLE,           // Sit quiet
  HOST_REVERBTEST,     // Generate REVERB for manufacturing test
  HOST_CAGCTEST,       // Set coarse receive gain for manufacturing test
  HOST_DGASP,          // send Dying Gasp messages through EOC channel
  HOST_GHSREQ,         // G.hs - FOR INTERNAL USE ONLY
  HOST_GHSMSG,         // G.hs - FOR INTERNAL USE ONLY
  HOST_GHS_SENDGALF,   // G.hs - FOR INTERNAL USE ONLY
  HOST_GHSEXIT,        // G.hs - FOR INTERNAL USE ONLY
  HOST_GHSMSG1,        // G.hs - FOR INTERNAL USE ONLY
  HOST_HYBRID,         // Enable/Disable automatic hybrid switch
  HOST_RJ11SELECT,     // RJ11 inner/outer pair select
  HOST_DIGITAL_MEM,    // Digital Diags: run external memory tests
  HOST_TXREVERB,       // AFE Diags: TX path Reverb
  HOST_TXMEDLEY,       // AFE Diags: TX path Medley
  HOST_RXNOISEPOWER,   // AFE Diags: RX noise power
  HOST_ECPOWER,        // AFE Diags: RX eco power
  HOST_ALL_ADIAG,      // AFE Diags: all major analog diagnostic modes. Host is responsible to initiate each diagnostic sessions
  HOST_USER_ADIAG,     // AFE Diags: Host fills in analog diagnostic input data structure as specified and requests DSP to perform measurements as specified
  HOST_QUIT_ADIAG,     // AFE Diags: Host requests DSP to quit current diagnostic session. This is used for stopping the transmit REVERB/MEDLEY
  HOST_NO_CMD,         // All others - G.hs - FOR INTERNAL USE ONLY
  HOST_DSLSS_SHUTDOWN, // Host initiated DSLSS shutdown message
  HOST_SET_GENERIC,    // Set generic CO profile
  HOST_UNDO_GENERIC    // Set profile previous to Generic
};

enum
{
  DSP_IDLE,               // R_IDLE state entered
  DSP_ACTMON,             // R_ACTMON state entered
  DSP_TRAIN,              // R_TRAIN state entered
  DSP_ACTIVE,             // R_ACTIVE state entered
  DSP_XMITBITSWAP,        // Upstream bitswap complete - FOR INTERNAL USE ONLY
  DSP_RCVBITSWAP,         // Downstream bitswap complete - FOR INTERNAL USE ONLY
  DSP_RTDL,               // R_RTDL state entered - FOR INTERNAL USE ONLY
  DSP_RRTDLPKT,           // RTDL packet received - FOR INTERNAL USE ONLY
  DSP_XRTDLPKT,           // RTDL packet transmitted - FOR INTERNAL USE ONLY
  DSP_ERROR,              // Command rejected, wrong state for this command
  DSP_REVERBTEST,         // Manufacturing REVERB test mode entered
  DSP_CAGCTEST,           // Manufacturing receive gain test done
  DSP_OVERLAY_START,      // Notify host that page overlay has started - overlay number indicated by "tag"
  DSP_OVERLAY_END,        // Notify host that page overlay has ended   - overlay number indicated by "tag"
  DSP_CRATES1,            // CRATES1 message is valid and should be copied to host memory now
  DSP_SNR,                // SNR calculations are ready and should be copied to host memory now
  DSP_GHSMSG,             // G.hs - FOR INTERNAL USE ONLY
  DSP_RCVBITSWAP_TIMEOUT, // Acknowledge Message was not received within ~500 msec (26 Superframes).
  DSP_ATM_TC_SYNC,        // Indicates true TC sync on both the upstream and downstream. Phy layer ready for data xfer.
  DSP_ATM_NO_TC_SYNC,     // Indicates loss of sync on phy layer on either US or DS.
  DSP_HYBRID,             // DSP completed hybrid switch
  DSP_RJ11SELECT,         // DSP completed RJ11 inner/outer pair select
  DSP_INVALID_CMD,        // Manufacturing (Digital and AFE) diags: CMD received not recognized
  DSP_TEST_PASSED,        // Manufacturing diags: test passed
  DSP_TEST_FAILED,        // Manufacturing diags: test failed
  DSP_TXREVERB,           // Manufacturing AFE diags: Response to HOST_TXREVERB
  DSP_TXMEDLEY,           // Manufacturing AFE diags: Response to HOST_TXMEDLEY
  DSP_RXNOISEPOWER,       // Manufacturing AFE diags: Response to HOST_RXNOISEPOWER
  DSP_ECPOWER,            // Manufacturing AFE diags: Response to HOST_ECPOWER
  DSP_ALL_ADIAG,          // Manufacturing AFE diags: Response to HOST_ALL_ADIAG
  DSP_USER_ADIAG,         // Manufacturing AFE diags: Response to HOST_USER_ADIAG
  DSP_QUIT_ADIAG,         // Manufacturing AFE diags: Response to HOST_QUIT_ADIAG
  DSP_DGASP               // DSP Message to indicate dying gasp
};

static unsigned char analogNoTonesTestArray[64]=
  {
    0,0,0,0,0,0,0,0,  // Tones 01-08
    0,0,0,0,0,0,0,0,  // Tones 09-16
    0,0,0,0,0,0,0,0,  // Tones 17-24
    0,0,0,0,0,0,0,0,  // Tones 25-32
    0,0,0,0,0,0,0,0,  // Tones 33-40
    0,0,0,0,0,0,0,0,  // Tones 41-48
    0,0,0,0,0,0,0,0,  // Tones 49-56
    0,0,0,0,0,0,0,0   // Tones 57-64
  };

static unsigned char analogAllTonesTestArray[64]=
  {
    1,1,1,1,1,1,1,1,  // Tones 01-08
    1,1,1,1,1,1,1,1,  // Tones 09-16
    1,1,1,1,1,1,1,1,  // Tones 17-24
    1,1,1,1,1,1,1,1,  // Tones 25-32
    1,1,1,1,1,1,1,1,  // Tones 33-40
    1,1,1,1,1,1,1,1,  // Tones 41-48
    1,1,1,1,1,1,1,1,  // Tones 49-56
    1,1,1,1,1,1,1,1   // Tones 57-64
  };

static unsigned char analogEvenTonesTestArray[64]=
  {
    0,1,0,1,0,1,0,1,  // Tones 01-08
    0,1,0,1,0,1,0,1,  // Tones 09-16
    0,1,0,1,0,1,0,1,  // Tones 17-24
    0,1,0,1,0,1,0,1,  // Tones 25-32
    0,1,0,1,0,1,0,1,  // Tones 33-40
    0,1,0,1,0,1,0,1,  // Tones 41-48
    0,1,0,1,0,1,0,1,  // Tones 49-56
    0,1,0,1,0,1,0,1   // Tones 57-64
  };

static unsigned char analogOddTonesTestArray[64]=
  {
    1,0,1,0,1,0,1,0,  // Tones 01-08
    1,0,1,0,1,0,1,0,  // Tones 09-16
    1,0,1,0,1,0,1,0,  // Tones 17-24
    1,0,1,0,1,0,1,0,  // Tones 25-32
    1,0,1,0,1,0,1,0,  // Tones 33-40
    1,0,1,0,1,0,1,0,  // Tones 41-48
    1,0,1,0,1,0,1,0,  // Tones 49-56
    1,0,1,0,1,0,1,0   // Tones 57-64
  };

int shim_osGetCpuFrequency(void)
{
  return 150;
}

static void tn7dsl_adiag(int Test, unsigned char *missingTones)
{
  int rc,cmd, tag;

  rc = dslhal_diags_anlg_setMissingTones(pIhw,missingTones);
  if(rc)
  {
    printk(" failed to set Missing town\n");
    return;
  }

/*********** Start the actual test **********************/

 if(Test==0)
   {
   printk("TX REVERB Test\n");
   rc = dslhal_support_writeHostMailbox(pIhw, HOST_TXREVERB, 0, 0, 0);
   if (rc)
     {
     printk("HOST_TXREVERB failed\n");
     return;
     }

   }
 if(Test==1)
   {
     dgprintf(0,"TX MEDLEY Test\n");
     rc = dslhal_support_writeHostMailbox(pIhw, HOST_TXMEDLEY, 0, 0, 0);
     if (rc)
       return;
   }
 dgprintf(4,"dslhal_diags_anlg_testA() done\n");
 return;
}


static void tn7dsl_diagnostic_test(char *data)
{
  if(!strcmp(data, "ADIAGRALL"))
  {
    printk("TX Reverb All tone\n");
    tn7dsl_adiag(0,analogAllTonesTestArray);
    return;
  }
  if(!strcmp(data, "ADIAGRNONE"))
  {
    printk("TX Reverb No tone\n");
    tn7dsl_adiag(0,analogNoTonesTestArray);
    return;
  }
  if(!strcmp(data, "ADIAGREVEN"))
  {
    printk("TX Reverb Even tone\n");
    tn7dsl_adiag(0,analogEvenTonesTestArray);
    return;
  }
  if(!strcmp(data, "ADIAGRODD"))
  {
    printk("TX Reverb Odd tone\n");
    tn7dsl_adiag(0,analogOddTonesTestArray);
    return;
  }
  if(!strcmp(data, "ADIAGMALL"))
  {
    printk("TX Mdelay All tone\n");
    tn7dsl_adiag(1,analogAllTonesTestArray);
    return;
  }
  if(!strcmp(data, "ADIAGMNONE"))
  {
    printk("TX Mdelay No tone\n");
    tn7dsl_adiag(1,analogNoTonesTestArray);
    return;
  }
  if(!strcmp(data, "ADIAGMEVEN"))
  {
    printk("TX Mdelay Even tone\n");
    tn7dsl_adiag(1,analogEvenTonesTestArray);
    return;
  }
  if(!strcmp(data, "ADIAGMODD"))
  {
    printk("TX Mdelay Odd tone\n");
    tn7dsl_adiag(1,analogOddTonesTestArray);
    return;
  }
  if(!strcmp(data, "ADIAGQUIET"))
  {
    dslhal_api_sendIdle(pIhw);
    return;
  }
  if(!strncmp(data, "ADIAGRN", 7))
  {
    char tones[64], tmp[4];
    int  nth, i;

    tmp[0]=data[7];
    tmp[1]=data[8];
    tmp[2]=data[9];

    nth = os_atoi(tmp);

    for(i=0;i<64;i++)
    {
      if(((i+1)% nth)==0)
      {
        tones[i]=0;
      }
      else
      {
        tones[i]=1;
      }
    }
    printk("TX Reverb with %dth tones missing.\n", nth);
    tn7dsl_adiag(0,tones);
    return;
  }
  if(!strncmp(data, "ADIAGMN", 7))
  {
    char tones[64], tmp[4];
    int  nth, i;

    tmp[0]=data[7];
    tmp[1]=data[8];
    tmp[2]=data[9];
    nth = os_atoi(tmp);

    for(i=0;i<64;i++)
    {
      if(((i+1)% nth)==0)
      {
        tones[i]=0;
      }
      else
      {
        tones[i]=1;
      }
    }
    printk("TX Mdelay with %dth tones missing.\n", nth);
    tn7dsl_adiag(1,tones);
    return;
  }


}

#endif

static void tn7dsl_chng_modulation(void* data)
{
  unsigned char tmp1[64], tmp2[64], *p;
  int k, m;
 
  if(!strcmp(data, "trellison"))
  {
    //printk("set Trellis on\n");
    dslhal_api_setTrellisFlag(pIhw, 1);
    trellis = 1;
    dslhal_api_sendQuiet(pIhw);
    return;
  }
  if(!strcmp(data, "trellisoff"))
  {
    //printk("set Trellis off\n");
    dslhal_api_setTrellisFlag(pIhw, 0);
    trellis = 0;
    dslhal_api_sendQuiet(pIhw);
    return;
  }

  // pair selection
  if(!strcmp(data, "innerpair"))
  {
    //printk("set inner pair\n");
    dslhal_api_selectInnerOuterPair(pIhw, 0);
    dslhal_api_sendQuiet(pIhw);
    return;
  }
  if(!strcmp(data, "outerpair"))
  {
    //printk("set outer pair\n");
    dslhal_api_selectInnerOuterPair(pIhw, 1);
    dslhal_api_sendQuiet(pIhw);
    return;
  }

  //@Send modem to Idle.
  if(!strcmp(data, "idle"))
  {
    //printk("send idle\n");
    dslhal_api_sendIdle(pIhw);
    return;
  }
  //@Send modem to retrain.
  if(!strcmp(data, "retrain"))
  {
    //printk("retrain\n");
    dslhal_api_sendQuiet(pIhw);
    return;
  }

  //fine gain
  if(!strcmp(data, "gainctlon"))
  {
    //printk("gain control on\n");
    tn7dsl_ctrl_fineGain(1);
    dslhal_api_sendQuiet(pIhw);
    return;
  }
  if(!strcmp(data, "gainctloff"))
  {
    //printk("gain control off\n");
    tn7dsl_ctrl_fineGain(0);
    dslhal_api_sendQuiet(pIhw);
    return;
  }

  //dgas polarity
  if(!strcmp(data, "gaspolarityon"))
  {
    printk("gaspolarityon\n");
    dslhal_api_configureDgaspLpr(pIhw, 1, 1);
    dslhal_api_configureDgaspLpr(pIhw, 0, 1);
    return;
  }
  if(!strcmp(data, "gaspolarityoff"))
  {
    printk("gaspolarityoff\n");
    dslhal_api_configureDgaspLpr(pIhw, 1, 1);
    dslhal_api_configureDgaspLpr(pIhw, 0, 0);
    return;
  }

  if(!strcmp(data, "losalarmon"))
  {
    printk("losalarmon\n");
    dslhal_api_disableLosAlarm(pIhw, 0);
    return;
  }

  if(!strcmp(data, "losalarmoff"))
  {
    printk("losalarmoff\n");
    dslhal_api_disableLosAlarm(pIhw, 1);
    return;
  }

  if(!strcmp(data, "bitswapon"))
  {
    int offset[2] = {33, 0};

//  printk("set bitswap on\n");
    tn7dsl_generic_read(2, offset);
    dslReg |= dslhal_support_byteSwap32(0x00000001);
    tn7dsl_generic_write(2, offset, dslReg);    // enable Tx bitswap

    offset[0] = 32;       // olrDspRx_p
    tn7dsl_generic_read(2, offset);
    dslReg |= dslhal_support_byteSwap32(0x00000001);
    tn7dsl_generic_write(2, offset, dslReg);    // enable Rx bitswap

    dslhal_api_sendQuiet(pIhw);
    return;
  }

  if(!strcmp(data, "bitswapoff"))
  {
    int offset[2] = {33, 0};

//  printk("set bitswap off\n");
    tn7dsl_generic_read(2, offset);  // olrDspTx_p
    dslReg &= dslhal_support_byteSwap32(0xffffff00);
    tn7dsl_generic_write(2, offset, dslReg);   // disable Tx bitswap

    offset[0] = 32;       // olrDspRx_p
    dslReg &= dslhal_support_byteSwap32(0xffffff00);
    tn7dsl_generic_write(2, offset, dslReg);   // disable Rx bitswap

    dslhal_api_sendQuiet(pIhw);
    return;
  }

  if(!strcmp(data, "rmsgsra"))
  {
     int offset[2] = {5, 1};
     unsigned char rMsgsRA[12];

     dslhal_api_dspInterfaceRead(pIhw, (unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset, (unsigned char *)&rMsgsRA[0], 12);

     printk("rMsgsRA = 0x%02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
          rMsgsRA[0], rMsgsRA[1], rMsgsRA[2], rMsgsRA[3],
          rMsgsRA[4], rMsgsRA[5], rMsgsRA[6], rMsgsRA[7],
          rMsgsRA[8], rMsgsRA[9], rMsgsRA[10], rMsgsRA[11] );
  }

  p=(unsigned char *)data;

  // fine gain
  for(k=0;k<5;k++)
  {
    tmp1[k]=p[k];
  }
  tmp1[k]=0;
  if(!strcmp(tmp1, "gainv"))
  {
    for(m=0;m<4;m++)
    {
      tmp2[m]=p[m+k];
    }
    tmp2[m]=0;
    //printk("name =%s, %d\n", tmp1, os_atoh(tmp2));
    tn7dsl_set_fineGainValue(os_atoh(tmp2));
    return;
  }

  // interleave depth
  for(k=0;k<6;k++)
  {
    tmp1[k]=p[k];
  }
  tmp1[k]=0;
  if(!strcmp(tmp1, "idepth"))
  {
    for(m=0;m<1;m++)
    {
      tmp2[m]=p[m+k];
    }
    tmp2[m]=0;
    printk("name =%s, %d\n", tmp1, os_atoi(tmp2));
    dslhal_api_setMaxInterleaverDepth(pIhw, os_atoi(tmp2));
    dslhal_api_sendQuiet(pIhw);
    return;
  }


  // carrier per bit
  for(k=0;k<13;k++)
  {
    tmp1[k]=p[k];
  }
  tmp1[k]=0;
  if(!strcmp(tmp1, "carrierperbit"))
  {
    for(m=0;m<4;m++)
    {
      tmp2[m]=p[m+k];
    }
    tmp2[m]=0;
    //printk("name =%s, %d\n", tmp1, os_atoh(tmp2));
    dslhal_api_setMaxBitsPerCarrierUpstream(pIhw, os_atoh(tmp2));
    dslhal_api_sendQuiet(pIhw);
    return;
  }

  p=(unsigned char *)data;
  if(p[0]==0x58 || p[0]==0x78) //"X" or "x"
  {
    unsigned int dAddr;
    //printk("main address = 0x%x\n", pIhw->pmainAddr);
    for(k=0;k<8;k++)
    {
      tmp1[k]=p[k+1];
    }
    tmp1[k]=0;
    dAddr=os_atoh(tmp1);
    printk("address =0x%x\n", os_atoh(tmp1));
    if(p[k+1]==0x4c || p[k+1]==0x6c)
    {
      unsigned int dVal;
      k++;
      for(m=0;m<8;m++)
      {
        tmp2[m]=p[m+k+1];
      }
      tmp2[m]=0;
      dVal=os_atoh(tmp2);
      printk("writing value =0x%x\n",dVal);
      dslhal_support_blockWrite((void *)&(dVal), (void *)dAddr, 4);
    }
    else
    {
      dslhal_support_blockRead((void *)dAddr, &dslReg, 4);
      printk("dslReg=0x%x\n", dslReg);
    }
  }

  p=(unsigned char *)data;
  if(p[0]==0x59 || p[0]==0x79) //"Y" or "y"
  {
    for(k=0;k<8;k++)
    {
      tmp1[k]=p[k+1];
      if(p[k+1]== 0)
        break;
    }
    tmp1[k]=0;
    printk("write mail box cmd =%d\n", os_atoi(tmp1));
    dslhal_support_writeHostMailbox(pIhw, os_atoi(tmp1), 0, 0, 0);
  }


#ifdef ADIAG
  tn7dsl_diagnostic_test(data);
#endif

  /* The command is to set the modulation. */
  if(!tn7dsl_set_modulation(data, TRUE))
  {
      dslhal_api_sendQuiet(pIhw);
      return;
  }

}

static unsigned int tn7dsl_set_modulation(void* data, int flag)
{
    int retval = TRUE;
    int train_mode = -1;
    int size;
    char *cp = (char *)data;

    dgprintf(4,"tn7dsl_set_modulation\n");

    if(!strcmp(data, "T1413"))
    {
        train_mode = T1413_MODE;
    }
    else if(!strcmp(data, "GDMT"))
    {
        train_mode = GDMT_MODE;
    }
    else if(!strcmp(data, "GLITE"))
    {
        train_mode = GLITE_MODE;
    }
    else if(!strcmp(data, "MMODE"))
    {
        train_mode = MULTI_MODE;
    }
    else if(!strcmp(data, "NMODE"))
    {
        train_mode = NO_MODE;
    }
    else if(!strcmp(data, "AD2MOD"))
    {
        train_mode = ADSL2_MODE;
    }
    else if(!strcmp(data, "AD2DEL"))
    {
        train_mode = ADSL2_DELT;
    }
    else if(!strcmp(data, "A2PMOD"))
    {
        train_mode = ADSL2_PLUS;
    }
    else if(!strcmp(data, "A2PDEL"))
    {
        train_mode = ADSL2_PLUS_DELT;
    }
    else
        {
        /* Since we don't want to compare the nul terminator */
        size = strlen("tmode");
        
        if(!strncmp(cp, "tmode", size))
        {
            cp += size;

           /* Read the specified mode, which has to be a hex value */
           if((*cp == '0') && ((*(cp + 1) == 'x') || (*(cp + 1) == 'X')))
            {
               //sscanf(cp + 2, "%x",&train_mode);
               train_mode = os_atoh(cp + 2);
            }
            else
            {
                //sscanf(cp, "%x",&train_mode);
                train_mode = os_atoh(cp);
            }
        }
        else
        {
            /* CQ9605: The new default command format for changing the modulation. */
            if((*cp == '0') && ((*(cp + 1) == 'x') || (*(cp + 1) == 'X')))
            {
                //sscanf(cp + 2, "%x",&train_mode);
                train_mode = os_atoh(cp + 2);
            }
            else if(!flag)
            {
               /* For backward compatibility for modulation */       
              //sscanf(cp, "%d",&train_mode);
              train_mode = os_atoi(cp);
        
            }
        }
    }  

    if(train_mode >= 0)
    {
        printk("%s : Setting mode to %#x\n", __FUNCTION__, train_mode);
        retval = dslhal_api_setTrainingMode(pIhw, train_mode);
    }

    return retval;
}

#ifdef CONFIG_LED_MODULE
static void tn7dsl_led_on(unsigned long parm)
{
  dslhal_api_configureLed(pIhw,parm, 0);
}


static void tn7dsl_led_off(unsigned long parm)
{
  dslhal_api_configureLed(pIhw,parm, 1);
}

static void tn7dsl_led_init(unsigned long parm)
{
  dslhal_api_configureLed(pIhw,parm, 2);
}
#endif

static void tn7dsl_register_dslss_led(void)
{
#ifdef CONFIG_LED_MODULE

    // register led0 with led module
    ledreg[0].param = 2;
    ledreg[0].init = (void *)tn7dsl_led_init;
    ledreg[0].onfunc = (void *)tn7dsl_led_on;
    ledreg[0].offfunc = (void *)tn7dsl_led_off;
    register_led_drv(LED_NUM_1, &ledreg[0]);

    // register led1 output with led module
    ledreg[1].param = 1;
    ledreg[1].init = (void *)tn7dsl_led_init;
    ledreg[1].onfunc = (void *)tn7dsl_led_on;
    ledreg[1].offfunc = (void *)tn7dsl_led_off;
    register_led_drv(LED_NUM_2, &ledreg[1]);
#endif
}

static void tn7dsl_ctrl_fineGain(int value)
{
  int offset[2]={25,2},fineGain_f=0;

  //printk("gain control=%d\n", value);
  dslhal_api_dspInterfaceRead(pIhw,(unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset, (unsigned char *)&fineGain_f, 4);
  if(value == 1)
  {
    fineGain_f |= dslhal_support_byteSwap32(0x01000000);
  }
  else
  {
    fineGain_f &= dslhal_support_byteSwap32(0x00ffffff);
  }
  dslhal_api_dspInterfaceWrite(pIhw,(unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset, (unsigned char *)&fineGain_f, 4);
}

static void tn7dsl_set_fineGainValue(int value)
{
  int offset[2]={25,4};
  int fineGainVal=0;

  //printk("gain value =%d\n", value);
  if(value)
   {
    dslhal_api_dspInterfaceRead(pIhw,(unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset, (unsigned char *)&fineGainVal, 4);
    fineGainVal &= dslhal_support_byteSwap32(0xffff0000);
    fineGainVal |= dslhal_support_byteSwap32((value & 0xffff));
    dslhal_api_dspInterfaceWrite(pIhw,(unsigned int)pIhw->pmainAddr, 2, (unsigned int *)&offset, (unsigned char *)&fineGainVal, 4);
   }
}

int tn7dsl_generic_read(int offsetnum, int *offset)
{

  dslhal_api_dspInterfaceRead(pIhw,(unsigned int)pIhw->pmainAddr, offsetnum, (unsigned int *)offset, (unsigned char *)&dslReg, 4);
  //printk("value = %d\n", dslReg);
  return 0;
}

int tn7dsl_generic_write( int offsetnum, int *offset, int data)
{
  dslhal_api_dspInterfaceWrite(pIhw,(unsigned int)pIhw->pmainAddr, offsetnum, (unsigned int *)offset, (unsigned char *)&data, 4);
  return 0;
}

int tn7dsl_proc_write_stats (struct file *fp, const char * buf, unsigned long count, void * data)
{
        char local_buf[31];
    int  ret_val = 0;

        if(count > 30)
        {
                printk("Error : Buffer Overflow\n");
                printk("Use \"echo 0> avsar_modem_stats \" to reset the statistics\n");
                return -1;
        }

        copy_from_user(local_buf,buf,count);
        local_buf[count-1]='\0'; /* Ignoring last \n char */
        ret_val = count;

        if(strcmp("0",local_buf)==0)
        {
            /* Valid command */
            printk("Resetting AAL5 statistics.\n");
        sarStat.txPktCnt =0;
        sarStat.rxPktCnt =0;
        sarStat.txBytes = 0;
        sarStat.rxBytes = 0;
        sarStat.txErrors =0;
        sarStat.rxErrors =0;
        }
        else
        {
                printk("\nError: Unknown operation on DSL/AAL5 statistics\n");
                printk("Use \"echo 0 > avsar_modem_stats\" to reset the statistics\n");
                return -1;
        }

        return ret_val;
}


int tn7dsl_proc_train_mode_export(char* buf, char **start, off_t offset, int count,
                 int *eof, void *data)
{

    int len = 0;
    char *cp = buf + offset;
    int i = 0;
    int num_entries = sizeof(dsl_modes)/sizeof(dsl_modes[0]) ; 
    static int ctr = 0;

    if(len <= count)
    {    
        for (i = ctr; ((i < num_entries)&& (len <= count)) ; i++)
        {
            /* Write the current string only if we can fit it into the buffer */
            if((strlen(dsl_modes[i].mode_name) + 6 + len) <= count)
        {
                len += snprintf(cp+len, (count - len), "%s\t\t\t%#x\n", 
                                    dsl_modes[i].mode_name, dsl_modes[i].mode_value);
            }
            else
                break;
        }
    }

    /* Data was completely written */
    if (i >= num_entries)
    {  
    /* We are done with this */
        *eof = 1;
        ctr = 0;        
    }   
    else
    {
        /* We have not been able to write the complete data, and we have to 
           nul terminate the buffer.
        */
        *(cp + len) = '\0';
        
        /* Save the value of the counter for the next read 
            for the rest of the data.
        */
        ctr = i;        
    }

    return len;
}


