/** 
 *
 *  \file   dslhal_debug_functions.c
 *  \author Copyright (C) 2008 Infineon Technologies.
 *          All Rights Reserved
 *  \brief  AR7 Adam2 DSL HAL Debug functions
 *          This file contains the source code for the functions that call the underlying
 *          APIs of the DSL driver to provide debug infomation
 *
 *        Date          Version       Modified by     Comments
 *  \date 07/01/2008                  CZ              SMS00784491. Merged  
 
 *******************************************************************************************/
#include <adam2.h>
#include <sys/av1_reg.h>
#include <dsl_hal_api.h>
#include <dsl_hal_support.h>
#include "dslhal_app.h"
#include "dslhal_debug_functions.h"

/* The following variables are defined in the dslapp_main.c application */
extern tidsl_t *ptidsl;
extern MAINCONTROL *pmc;
extern TRAINMODE TrainModeTbl[];

// use "extern int rtraincnt;" if it defines in the other file */ 
int rtraincnt=0; 

/* local functions */
void internal_show_agc(void);
void internal_showtime_display_ADSL2p(int);
void internal_showtime_display(void);
unsigned short byteSwap16(unsigned short in16Bits);
unsigned   int byteSwap32(unsigned int in32Bits);

//
//@dgasp: send dying gasp
//
int dbgcmd_SendDyingGasp(int argc, unsigned char **argv)
{
  switch(argc) {
  case 1: // display current status
    printf("send dying gasp, bye!\n");
    dslhal_api_sendDgasp(ptidsl);
    break;

  case 2: // @dgasp
    if (askHelp())
    {
      printf("dgasp - Send dying gasp\n");
      printf("usage:\n");
      printf("  @dgasp  : send dying gasp\n");
      return 1; // ok
    }
  default:
    return 0; // err
  };  // switch

  return 1;
}


//@SNRpsds <0|1|2>
//  The result SNRpsds is in format 0&1 is in 8 bit entry as dfined in g.997.1
//  SNR(i) (dB) = (-32+(SNRpsds(i)/2))

int dbgcmd_SNRpsds(int argc, unsigned char **argv)
{
  int i;
  unsigned int offset[2]={7, 16};
  unsigned char SNRpsds[512];
  unsigned char *SNRpsds_p = &SNRpsds[0];

  printf("SNRpsds_p=%08x\n", PHYS_TO_K1(dslhal_support_readInternalOffset(ptidsl, 2, offset)));

  // call API instead of access internal buf directly
  if (dslhal_api_getSNRpsds(ptidsl, SNRpsds_p, 1) != DSLHAL_ERROR_NO_ERRORS)
  {
    printf("dslhal_api_getSNRpsds failed!\n");
    return 0;
  }

  dslhal_api_gatherStatistics(ptidsl); //We need to get the measurement time
  
  for (i=0; i<ptidsl->AppData.max_ds_tones; i++)
  {
    if (!(i%16))
      printf("\nSNRpsds(%3d):", i);
      printf(" %d", SNRpsds_p[i]);
  }
  printf("\nMeasurement Time: %d symbols\n\n", ptidsl->AppData.dsSNRMT);
  return 1;
}


// UR8_MERGE_START CQ11057 KCCHEN
void show_US_PMD_Message(tidsl_t * ptidsl, int disp_format)
{
  int i;
  CoPMDTestParams_t co_pmdtest_params;
  CoPMDTestParams_t *co_pmdtest_params_p = &co_pmdtest_params;

#if 0
  if (dslhal_api_getPMDTestus(ptidsl, co_pmdtest_params_p, 1) != DSLHAL_ERROR_NO_ERRORS)
  {
    printf("dslhal_api_getuspmd failed!\n");
    return ;
  }
#endif  

  printf("\nUSPMD\n");
  //HLOG
  for (i=0; i<ptidsl->AppData.max_us_tones; i++)
  {
    if (!(i%16))
      printf("\nHLOG(%3d):", i);
    if (disp_format==0)
      printf(" %d", co_pmdtest_params.TestParmCOHlogfMsg[i]);
    else
      printf(" %d", co_pmdtest_params.TestParmCOHlogfMsg[i]);
  }
  printf("\n");
  //UR8_MERGE_START_END CQ11709 Tim
  printf("Measurement Time: %d symbols\n", co_pmdtest_params.usHLOGMT);
  //QLN
  for (i=0; i<ptidsl->AppData.max_us_tones; i++)
  {
    if (!(i%16))
      printf("\nQLN(%3d):", i);
    if (disp_format==0)
      printf(" %d", co_pmdtest_params.TestParmCOQLNfMsg[i]);
    else
      printf(" %d", co_pmdtest_params.TestParmCOQLNfMsg[i]);
  }
  printf("\n");
  //UR8_MERGE_START_END CQ11709 Tim
  printf("Measurement Time: %d symbols\n", co_pmdtest_params.usQLNMT);
  //SNR
  for (i=0; i<ptidsl->AppData.max_us_tones; i++)
  {
    if (!(i%16))
      printf("\nSNR(%3d):", i);
    if (disp_format==0)
      printf(" %d", co_pmdtest_params.TestParmCOSNRfMsg[i]);
    else
      printf(" %d", co_pmdtest_params.TestParmCOSNRfMsg[i]);
  }
  printf("\n");
  //UR8_MERGE_START_END CQ11709 Tim
  printf("Measurement Time: %d symbols\n\n", co_pmdtest_params.usSNRMT);
  printf("LATN=%d\n", co_pmdtest_params.co_latn);
  printf("SATN=%d\n", co_pmdtest_params.co_satn);
  printf("SNRM=%d\n", co_pmdtest_params.usMargin);
  printf("attndr=%d\n", co_pmdtest_params.co_attndr);
  printf("NearActatp=%d\n", co_pmdtest_params.co_near_actatp);
  printf("FarActatp=%d\n", co_pmdtest_params.co_far_actatp);
}

//
//@pmdus <0|1>
//
int dbgcmd_pmdus(int argc, unsigned char **argv)
{
  int disp_format=0;
  
  show_US_PMD_Message(ptidsl, disp_format);
  
  return 1;
}
// UR8_MERGE_END CQ11057 KCCHEN
//
//@QLNpsds <0|1>
//
int dbgcmd_QLNpsds(int argc, unsigned char **argv)
{
  int i;
  int disp_format=0;
  unsigned int offset[2]={7, 15};
  unsigned char QLNpsds[512];
  unsigned char *QLNpsds_p = &QLNpsds[0];

  if (argc==2)
  {
    if (askHelp())
    {
      printf("@qlnpsds - show Quiet Line Noise DS per subcarrier\n");
      printf("usage: @qlnpsds <format:0|1>\n");
      printf("   format: 0=decimal (default), 1=hex\n");
      printf("Note:  only avail for ADSL2\2+ mode\n");
      return 1;
    }

    if (!GetNumber(argv[1], &disp_format))
      return 0; // err
    if (disp_format>1)
      return 0; // err
  }

  printf("QLNpsds_p=%08x\n", PHYS_TO_K1(dslhal_support_readInternalOffset(ptidsl, 2, offset)) );

  // call API instead of access internal buf directly
  if (dslhal_api_getQLNpsds(ptidsl, QLNpsds_p, 0) != DSLHAL_ERROR_NO_ERRORS)
  {
    printf("dslhal_api_getQLNpsds failed!\n");
    return 0;
  }

  //UR8_MERGE_START_END CQ11709 Tim
  dslhal_api_gatherStatistics(ptidsl); //We need to get the measurement time

  for (i=0; i<ptidsl->AppData.max_ds_tones; i++)
  {
    if (!(i%16))
      printf("\nQLNpsds(%3d):", i);
    if (disp_format==0)
      printf(" %d", QLNpsds_p[i]);
    else
      printf(" %02x", QLNpsds_p[i]);
  }
  //putchar('\n');
  //UR8_MERGE_START_END CQ11709 Tim
  printf("\nMeasurement Time: %d symbols\n\n", ptidsl->AppData.dsQLNMT);
  return 1;
}

//  UR8_MERGE_START CQ10979   Jack Zhang
//
//@HLINpsds <0|1>
//
int dbgcmd_HLINpsds(int argc, unsigned char **argv)
{
  int i;
  int disp_format=0;
  unsigned int offset[2]={13, 11};
  short    HLNpsds[2*512];
  short   *HLINpsds_p = &HLNpsds[0];

  if (argc==2)
  {
    if (askHelp())
    {
      printf("@hlinpsds - show linear channel characteristics per subcarrier\n");
      printf("usage: @hlinpsds <format:0|1>\n");
      printf("   format: 0=decimal (default), 1=hex\n");
      printf("Note:  only avail for ADSL2\2+ mode\n");
      return 1;
    }

    if (!GetNumber(argv[1], &disp_format))
      return 0; // err
    if (disp_format>1)
      return 0; // err
  }

  printf("HLINpsds_p=%08x\n", PHYS_TO_K1(dslhal_support_readOffset(ptidsl, 2, offset)) );

  // call API instead of access internal buf directly
  if (dslhal_api_getHLINpsds(ptidsl, (unsigned char *)HLINpsds_p, 1) != DSLHAL_ERROR_NO_ERRORS)
  {
    printf("dslhal_api_getHLINpsds failed!\n");
    return 0;
  }

  for (i=0; i<ptidsl->AppData.max_ds_tones; i++)
  {
    if (!(i%8))
      printf("\nHLINpsds(%3d):", i);
    if (disp_format==0)
      printf(" (%d,%d)", HLINpsds_p[2*i], HLINpsds_p[2*i+1]);
    else
      printf(" (%02x,%02x)", HLINpsds_p[2*i], HLINpsds_p[2*i+1]);
  }
  putchar('\n');
  return 1;
}
//  UR8_MERGE_END   CQ10979*

//
//@HLOGpsds <0|1>
//
int dbgcmd_HLOGpsds(int argc, unsigned char **argv)
{
  int i;
  int disp_format=0;
  unsigned int offset[2]={7, 14};
  unsigned short HLOGpsds[512];
  unsigned short *HLOGpsds_p = &HLOGpsds[0];


  if (argc==2)
  {
    if (askHelp())
    {
      printf("@hlogpsds - show Hlog DS per subcarrier (per g.997.1)\n");
      printf("usage: @hlogpsds <format:0|1>\n");
      printf("   format: 0=decimal (default), 1=hex\n");
      printf("Note:  only avail for ADSL2\2+ mode\n");
      return 1;
    }

    if (!GetNumber(argv[1], &disp_format))
      return 0; // err

    if (disp_format>2)
      return 0; // err
  }

  
  printf("HLOGpsds_p=%08x\n", PHYS_TO_K1(dslhal_support_readInternalOffset(ptidsl, 2, offset)));

  // call API instead of access internal buf directly
  if (dslhal_api_getHLOGpsds(ptidsl, HLOGpsds_p, 0) != DSLHAL_ERROR_NO_ERRORS)
  {
    printf("dslhal_api_getHLOGpsds failed!\n");
    return 0;
  }

  //UR8_MERGE_START_END CQ11709 Tim
  dslhal_api_gatherStatistics(ptidsl);  //We need to get the measurement time

  for (i=0; i<ptidsl->AppData.max_ds_tones; i++)
  {
    if (!(i%16))
      printf("\nHLOGpsds(%3d):", i);
    if (disp_format==0)
      printf(" %d", HLOGpsds_p[i]);
    else
      printf(" %04x", HLOGpsds_p[i]);
  }
  //putchar('\n');
  //UR8_MERGE_START_END CQ11709 Tim
  printf("\nMeasurement Time: %d symbols\n\n", ptidsl->AppData.dsHlogMT);

  return 1;
}

//
//@quit : Quit the Dsl Debug App
//
int dbgcmd_quit(int argc, unsigned char **argv)
{
  if (argc==1)
  {
    printf("User Quiting...\n");
    // MySioFlush();
    if (pmc->dp_load) shutdownDsl();
    DSLHAL_REG32(REG_PRCR) = (PRCR_UART0 | PRCR_I2C | PRCR_GPIO);
    AllDone(0);

    return 1;
  }

  return 0;
}

//
// Display ADSL2\2+ status
//
void internal_showtime_display_ADSL2p(int flag)
{
  DDI_RX_MSG *recvMsg_p, recvMsg_s;
  DDI_TX_MSG *xmtMsg_p, xmtMsg_s;
  unsigned int recvMsgAddr_offset[2]={7, 17};
  unsigned int xmtMsgAddr_offset[2]={7, 18};
  unsigned int INP100;
  unsigned int Lp;

  if (pmc->DspVerS.major <6)
  {
    if (flag)
    {
      printf("This feature requires datapump V6.x+\n");
    }
    return;  // not supported
  }

  recvMsg_p = (DDI_RX_MSG *) dslhal_support_readInternalOffset(ptidsl, 2, recvMsgAddr_offset);
  dslhal_support_blockRead((void *) recvMsg_p, &recvMsg_s, sizeof(recvMsg_s));

  xmtMsg_p = (DDI_TX_MSG *) dslhal_support_readInternalOffset(ptidsl, 2, xmtMsgAddr_offset);
  dslhal_support_blockRead((void *) xmtMsg_p, &xmtMsg_s, sizeof(xmtMsg_s));

 //------------------------------------------------------------
  if (flag) {
    printf("Note: some of the parameters are for ADSL2\2+ only\n");
    printf("US: xmtMsgAddr_p=%08x\n", (unsigned int) xmtMsg_p );
  }

  Lp=dslhal_support_byteSwap16(xmtMsg_s.Lp);
  INP100=(4*xmtMsg_s.r*xmtMsg_s.D*100)/Lp; // INP=(4*R*D)/L
  printf("US: T=%d,K=%d,M=%d,R=%d,D=%d,L=%d,n=%d,INP=%d.%02d\n",
    xmtMsg_s.Tp,
    xmtMsg_s.k,
    xmtMsg_s.s,
    xmtMsg_s.r,
    xmtMsg_s.D,
    Lp,
    //UR8_MERGE_START_END CQ11744 KCCHEN
    (xmtMsg_s.s*xmtMsg_s.k+xmtMsg_s.r),
    INP100/100, INP100%100);

  //--------------------------------------------------------
  if (flag)
    printf("DS: recvMsgAddr_p=%08x\n", (unsigned int) recvMsg_p );

  Lp=dslhal_support_byteSwap16(recvMsg_s.Lp);
  recvMsg_s.D=dslhal_support_byteSwap16(recvMsg_s.D);
  INP100=(4*recvMsg_s.r*recvMsg_s.D*100)/Lp;  // INP=(4*R*D)/L
  printf("DS: T=%d,K=%d,M=%d,R=%d,D=%d,L=%d,n=%d,INP=%d.%02d\n",
    recvMsg_s.Tp,
    dslhal_support_byteSwap16(recvMsg_s.k),
    //UR8_MERGE_START_END CQ11744 KCCHEN
    // N1 = MK+R. Only Rx has N1 field.
    recvMsg_s.s, recvMsg_s.r, recvMsg_s.D, Lp, dslhal_support_byteSwap16(recvMsg_s.N1),
    INP100/100, INP100%100);
    
}

// This function is invoked right after enter showtime.
// The detailes of what info be didsplayed is controlled by
// MainControl.showTimeDispControl, which is set by @showtime
// cmd.
//
void internal_showtime_display(void)
{
static struct {
  unsigned int mask;
  void (*func)(void);
} showtime_disp_table[] = {
    {SHOWTIME_DISP_BITALLOC_DS, disp_bitalloc_ds},
    {SHOWTIME_DISP_BITALLOC_US, disp_bitalloc_us},
    {SHOWTIME_DISP_DSPERROR,    disp_dsp_error},
    {SHOWTIME_DISP_DSMARGIN,    disp_margin_ds},
};
  int i;

  for (i=0; i<NUM_ELMS(showtime_disp_table); i++)
  {
    if (pmc->showTimeDispControl & showtime_disp_table[i].mask)
      showtime_disp_table[i].func();
  }
}


//
//@showtime : display showtime info & set showtime display control
//
int dbgcmd_ShowTimeInfo(int argc, unsigned char **argv)
{

  switch (argc) {
  case 1: //@showtime: simply display showtime info
    printf("showTimeDisp=0x%x\n", pmc->showTimeDispControl);
    dslhal_api_gatherStatistics(ptidsl);

    printf("\nDSLShowTime %s(0x%x) (%s)\n",
      TrainModeFromNum(ptidsl->AppData.TrainedMode),
      ptidsl->AppData.TrainedMode,
      AnnexFromNum(ptidsl->AppData.annex_selected));

    display_ShowTimeInfo (ptidsl);
    break;

  default:
    return 0; // err
  };

  return 1;
}

void internal_show_agc(void)
{
  int offset[]={4};
  DEV_HOST_agcSetting_t *agcSetting_p, agcSetting;

  agcSetting_p=(DEV_HOST_agcSetting_t *) dslhal_support_readInternalOffset(ptidsl, 1, offset);
  dslhal_support_blockRead((void*) agcSetting_p, &agcSetting, sizeof(agcSetting));

  printf("rxPga=%d,%d,%d, anlgEq=%d\n",
    agcSetting.rxPga1, agcSetting.rxPga2, agcSetting.rxPga3, agcSetting.anlgEq);
}

//
//@pcb <0|1>: show/disable/enable PowerCutback
//
int dbgcmd_setPowerCutBack(int argc, unsigned char **argv)
{
  int offset[]={25};
  DEV_HOST_modemEnvPublic_t *modemEnvPublic_p;
  unsigned char pcbEnabled;
  unsigned int  pcbOnOff;

  switch (argc) {
  case 1: //@pcb :show current setting
    modemEnvPublic_p=(DEV_HOST_modemEnvPublic_t *) dslhal_support_readOffset(ptidsl, 1, offset);
    printf("addr(pcbEnabled)=%08x\n", (unsigned int) &modemEnvPublic_p->pcbEnabled);
    dslhal_support_blockRead((void*) &modemEnvPublic_p->pcbEnabled,
      &pcbEnabled, sizeof(modemEnvPublic_p->pcbEnabled));
    printf("pcb=%d\n", pcbEnabled);
    break;

  case 2:
    if (askHelp())
    {
      printf("@pcb - display power cut back messages\n");
      printf("usage:\n");
      printf("  @pcb : show current setting\n");
      printf("  @pcb <0|1>: disable/enable power cutback\n");
      return 1;
    }

    if (!GetNumber(argv[1], &pcbOnOff))
      return 0; // err
    if ((pcbOnOff !=0) && (pcbOnOff!=1))
      return 0; // err
    printf("pcb<-%d\n", pcbOnOff);
    dslhal_advcfg_onOffPcb(ptidsl, pcbOnOff);
    break;

  default:
    return 0; // err
  };
  return 1;
}

//
//@snrd [n] [disp_fmt]
//  n: snr buffer:0
//  disp_fmt: 0: Raw(decimal)
//
int dbgcmd_SNR_Ds (int argc, unsigned char **argv)
{
  unsigned short *pSnrBuf;
  int i;
  int num_ds_bins;
  signed short snr_ds;

  num_ds_bins= (ptidsl->AppData.TrainedMode& ADSL2PLUS_MASKS) ? 512: 256;

  pSnrBuf=(unsigned short *) &(ptidsl->AppData.rxSnrPerBin0[0]); 
    
    for(i=0; i<num_ds_bins; i++)
    {
      if(!(i%16)) printf("\nSNRD(%3d):", i);
      snr_ds=(short) byteSwap16(pSnrBuf[i]);    
      printf(" %d", snr_ds);     
    }
    printf("\n");

  return 1;
}

void disp_bitalloc_us(void)
{
#define NUM_US_BINS  64
  unsigned char usBits[NUM_US_BINS];
  short usGains[NUM_US_BINS];
  int num_us_bins;
  int i, total_bits;

  num_us_bins = dslhal_support_getMaxUsTones(ptidsl);   // hardcode us bins for now
  dslhal_api_gatherStatistics(ptidsl);
  memset(usBits,0, sizeof(usBits));
  memset(usGains,0, sizeof(usGains));

  for(i=1; i<num_us_bins; i++)
  {
    usBits[i]=((ptidsl->AppData.cBng[(i-1)*2])&0xf);
    usGains[i]= ((ptidsl->AppData.cBng[(2*i)-1]) <<4);
    usGains[i]+= (((ptidsl->AppData.cBng[(i-1)*2])&0xf0) >>4);
  }
  for (total_bits=i=0;  i<num_us_bins;  i++)
  {
    if (!(i%8)) printf("\nBAU(%3d):", i);
    if (pmc->bitalloc_us_dispformat==1)
      printf(" %02x", usBits[i]);
    else
      printf(" %02d", usBits[i]);
    total_bits+=usBits[i];
  }
  printf("\nBAU: Total %d bits\n", total_bits);
  printf("\nBAU: US fine gain table");
  for (i=0;  i<num_us_bins;  i++)
  {
    if (!(i%8)) printf("\nBAU(%3d):", i);
    if (pmc->bitalloc_us_dispformat==1)
      printf(" %03x", usGains[i]);
    else
      printf(" %d", usGains[i]);
  }

  printf("\n");
}

//
//@mard: dump downstream margin
//
int dbgcmd_Margin_Ds(int argc, unsigned char **argv)
{
  int disp_format=0;  // default, decimal

  switch (argc) {
  case 2: // @mard <0|1>
    if (askHelp())
    {
      printf("@mard - show DS margin table (added by YT 2005/11/4)\n");
      printf("usage: @mard <0|1>\n");
      printf("  0:decimal (default), 1:hex\n");
      return 1;
    }
    if (!GetNumber(argv[1], &disp_format))
      return 0;  // error

    if (disp_format>1) return 0;   // error
    pmc->margin_ds_dispformat= disp_format;
    break;
  default:
    break;
  }

  disp_margin_ds();
  return 1;
}

void disp_margin_ds(void)
{
  signed short dsMargins[512];
  int i; 
  int num_ds_bins;

  dslhal_api_gatherStatistics(ptidsl);
  memset(dsMargins,0, sizeof(dsMargins));

  num_ds_bins= dslhal_support_getMaxDsTones(ptidsl);

  for(i=1; i<num_ds_bins; i++)
  { 
    dsMargins[i]=(ptidsl->AppData.marginTblDstrm[i]/2);
  }

  for (i=0;  i<num_ds_bins;  i++)
  {
    if (!(i%16)) printf("\nMARD(%3d):", i);

    if (pmc->margin_ds_dispformat==1)
      printf(" %02x", dsMargins[i]);
    else
      printf(" %3d", dsMargins[i]);
  }

  printf("\n");
}

void disp_bitalloc_ds(void)
{
  unsigned char dsBits[512];
  short dsGains[512];
  int i, total_bits;
  int num_ds_bins;

  dslhal_api_gatherStatistics(ptidsl);
  memset(dsBits,0, sizeof(dsBits));
  memset(dsGains,0, sizeof(dsGains));

  num_ds_bins= dslhal_support_getMaxDsTones(ptidsl);

  for(i=1; i<num_ds_bins; i++)
  {
    dsBits[i]=((ptidsl->AppData.rBng[(i-1)*2])&0xf);
    dsGains[i]= ((ptidsl->AppData.rBng[(2*i)-1]) <<4);
    dsGains[i]+= (((ptidsl->AppData.rBng[(i-1)*2])&0xf0) >>4);
  }

  for (total_bits=i=0;  i<num_ds_bins;  i++)
  {
    if (!(i%16)) printf("\nBAD(%3d):", i);
    if (pmc->bitalloc_ds_dispformat==1)
      printf(" %02x", dsBits[i]);
    else
      printf(" %2d", dsBits[i]);
    total_bits+= dsBits[i];
  }
  printf("\nBAD: Total %d bits\n", total_bits);

  printf("\nBAD: DS fine gain table");
  for (i=0;  i<num_ds_bins;  i++)
  {
    if (!(i%16)) printf("\nBAD(%3d):", i);
    if (pmc->bitalloc_ds_dispformat==1)
      printf(" %03x",dsGains[i]);
    else
      printf(" %3d",dsGains[i]);
  }
  printf("\n");
}

//
//@bad <0|1>: Bit Allocation DS
//
int dbgcmd_BitAllocDs (int argc, unsigned char **argv)
{
  int disp_format=0;  // default, decimal

  switch (argc) {
  case 2: // @bad <0|1>
    if (askHelp())
    {
      printf("@bad - show DS bit allocation table & fine gain table\n");
      printf("usage: @bad <0|1>\n");
      printf("  0:decimal (default), 1:hex\n");
      return 1;
    }
    if (!GetNumber(argv[1], &disp_format))
      return 0;  // error

    if (disp_format>1) return 0;   // error
    pmc->bitalloc_ds_dispformat= disp_format;
    break;
  default:
    break;
  }

  disp_bitalloc_ds();
  return 1;
}

//
// @bau <0|1>: Bit Allocation US
//
int dbgcmd_BitAllocUs (int argc, unsigned char **argv)
{
  int disp_format;

  switch (argc) {
  case 2: // @bau <0|1>
    if (askHelp())
    {
      printf("@bau - show US bit allocation table & fine gain table\n");
      printf("usage: @bau <0|1>\n");
      printf("  0:decimal (default), 1:hex\n");
      return 1;
    }
    if (!GetNumber(argv[1], &disp_format))
      return 0;  // error

    if (disp_format>1) return 0;   // error
    pmc->bitalloc_us_dispformat= disp_format;
    break;
  default:
    break;
  }

  disp_bitalloc_us();
  return 1;
}

//
//@dstones : TuenOn/Off DS Tones
//
//Note:
//  The effect of disabling ds tones is non-accumulative.
//  i.e. the setting from previous time won't affact the setting of this time.
//
int dbgcmd_dsTonesOnOff(int argc, unsigned char **argv)
{
  int offset[2] = {4, 32};
  DEV_HOST_oamWrNegoParaDef_t *oamWrNego_p, oamWrNego;
  unsigned int missing_tone;  //, num_missing_tones;
  int i;
  unsigned short tmp16;
  unsigned short MissDsTones[64];

  oamWrNego_p= (DEV_HOST_oamWrNegoParaDef_t *) dslhal_support_readOffset(ptidsl, 1, (unsigned int *) &offset);
  dslhal_support_blockRead((void *) oamWrNego_p, &oamWrNego, sizeof(DEV_HOST_oamWrNegoParaDef_t));

  switch (argc) {
  case 1:
    printf("dsToneTurnoff_f=%08x\n", (unsigned int) oamWrNego.dsToneTurnoff_f);
    printf("missingToneDsAddr=%08x\n",
      (unsigned int) dslhal_support_byteSwap32(oamWrNego.missingToneDsAddr));
    break;

  case 2:
    if (askHelp())
    {
      printf("@dstones - disable DS tones for training\n");
      printf("usage: @dstones <tone1> <tone2> ..<tone n>\n");
      printf("Note: <tone number> : 16 bit number (decimal or 0x.. hex) can be a individual tone number\n");
      printf("     or a range of tones.\n");
      printf("  To specify the range of the tones, use MSB 01 for range start, MSB10 for range end\n");
      printf("  Can use @bad after showtime to confirm the tones are disabled.\n");
      printf("eg. @dstones 123 0x4080 0x8120 0x140\n");
      printf(" Will disable tone 123, 128 (0x80)-288 (0x120), and tone 320 (0x140)\n");
      return 1;
    }

  default:
    // set DS missing tones
    for (i=0, argv++;  i<(argc-1);  i++, argv++)
    {
      if (!GetNumber(argv[0], &missing_tone))
        return 0;  // error
      MissDsTones[i] = (unsigned short) missing_tone;
    }
    MissDsTones[i]=0;

    printf("missing_tones:");
    for (i=0; MissDsTones[i]!=0;  i++)
    {
      tmp16 = MissDsTones[i];

      if (i>32) return 0;  // error
      printf("0x%04x(%d)", tmp16, tmp16 & ~0xc000);
      if (tmp16 & 0x4000) // range start
        printf("-");
      else printf(",");
    }

    printf("\n");
    dslhal_advcfg_configDsTones(ptidsl, MissDsTones);
    break;

  } // switch

  return 1;
}

//
//@bswp <ds:0|1> <us:0|1>
//@bswp <1|0>
//
int dbgcmd_BitSwapOnOff (int argc, unsigned char **argv)
{
  unsigned int bswp_ds, bswp_us;
  unsigned int dslReg;
  int offset_ds[2] = {32, 0};
  int offset_us[2] = {33, 0};

  switch (argc) {
  case 2: // @bswp <0|1>
    if (askHelp())
    {
      printf("@bswp - bitswap disable/enable control\n");
      printf("usage:\n");
      printf("  @bswp : show current status\n");
      printf("  @bswp <ds:0|1> <us:0|1>\n");
      printf("  eg. @bswp 0 1: disable ds bitswap, enable us bitswap\n");
      return 1;
    }

    if (!GetNumber(argv[1], &bswp_ds)) return 0;
    if ((bswp_ds!=0) && (bswp_ds!=1)) return 0;   // only 0|1 allowed
    bswp_us = bswp_ds;  // set same to both ds/us

    // intended flow thru
  case 3: // @bswp <ds:0|1> <us:0|1>
    if (argc==3)
    {
      if (!GetNumber(argv[1], &bswp_ds)) return 0;
      if (!GetNumber(argv[2], &bswp_us)) return 0;
    }
    dslhal_advcfg_onOffBitSwap(ptidsl,1,bswp_ds);  // Ds
    dslhal_advcfg_onOffBitSwap(ptidsl,0,bswp_us);  // Us
    printf("bswp_ds: %s, bswp_us:%s\n",
      bswp_ds? "on" : "off",
      bswp_us? "on" : "off");
    break;

  default: // show current setting
    dslReg= dslhal_support_readOffset(ptidsl, 2, (unsigned int *)&offset_ds);
    bswp_ds = dslReg & 0x000000ff;
    dslReg= dslhal_support_readOffset(ptidsl, 2, (unsigned int *)&offset_us);
    bswp_us = dslReg & 0x000000ff;
    printf("bswp_ds=%d, bswp_us=%d\n", bswp_ds? 1: 0, bswp_us ? 1 : 0);
    break;
  };   // switch

  return 1;
}

//
//@cd : set to Idle
//
int dbgcmd_setDslIdle(int argc, unsigned char **argv)
{
  if ((argc==2)&& askHelp())
  {
    printf("@cd - set to Idle\n");
    printf("usage: @cd\n");
    return 1;
  }

  printf("set to Idle\n");
  dslhal_api_sendIdle(ptidsl);
  rtraincnt=0; //UR8_MERGE_START_END CQ10602 ManjulaK
  return 1;
}

//
//@ca : set to Activate
//
int dbgcmd_setDslActivate(int argc, unsigned char **argv)
{
  if ((argc==2) && askHelp())
  {
    printf("@ca - set to Active\n");
    printf("usage: @ca\n");
    return 1;
  }

  printf("set to Active\n");
  dslhal_api_sendQuiet(ptidsl);
  return 1;
}

//
//@ce : set to Retrain
//
int dbgcmd_setDslRetrain(int argc, unsigned char **argv)
{
  if ((argc==2) && askHelp())
  {
    printf("@ce - set to Retrain\n");
    printf("usage: @ce\n");
    return 1;
  }

  printf("set to Retrain\n");
  dslhal_api_dslRetrain(ptidsl);
  return 1;
}

void disp_dsp_error(void)
{
  DEV_HOST_dspErrorCounts_t dspErrCount, *dspErrCount_p;
  CTRL_MAIN_exceptionCounts_t exceptionCounts;
  unsigned int *p;
  int offset[]={6};
  unsigned int err_count, i;
  unsigned int RsNotDoneErrCnt;
  unsigned int AllOptionsFailCnt;
  unsigned int swdiTxErr, swdiRxErr;

  dspErrCount_p = (DEV_HOST_dspErrorCounts_t *) dslhal_support_readInternalOffset(ptidsl, 1, offset);

  printf("\ndspErrCount_p=%08x\n", (unsigned int) dspErrCount_p);

  // read Trained mode, don't need to swap since it's only 'char' (8 bit)
  dslhal_support_blockRead((void*) dspErrCount_p, &dspErrCount, sizeof(dspErrCount));

  for (i=0, p=(unsigned int* )&dspErrCount;  i<sizeof(dspErrCount)/4;  i++, p++)
    *p=dslhal_support_byteSwap32(*p);

  dslhal_support_blockRead((void*) dspErrCount.realTimeErrors_p, &exceptionCounts, sizeof(exceptionCounts));
  for (err_count=i=0, p=(unsigned int* )&exceptionCounts;  i<sizeof(exceptionCounts)/4;  i++, p++)
  {
    *p=dslhal_support_byteSwap32(*p);
    if (*p) err_count++;
  }  
  printf("   realTimeErrors=%d\n", exceptionCounts.realTimeErrors);  // counts number of real time errors 
}

//
//@eoc
//
int dbgcmd_EocInfo(int argc, unsigned char **argv)
{
  int offset[]= {11};
  DEV_HOST_eocVarDef_t eoc, *eoc_p;

  if (argc==2)
  {
    if (askHelp())
    {
      printf("@eoc - display EOC info\n");
      printf("usage: @eoc\n");
      return 1;
    }
  }


  eoc_p= (DEV_HOST_eocVarDef_t *) dslhal_support_readOffset(ptidsl, 1, offset);
  dslhal_support_blockRead((void *) eoc_p, &eoc, sizeof(DEV_HOST_eocVarDef_t));

  printf("eoc_p= %08x\n", (unsigned int) eoc_p);

  printf("eocReg4[32](%08x)=%02x %02x %02x %02x %02x %02x\n",
    (unsigned int) eoc_p->eocReg4,      // Host/Dsp Write, vendor specific EOC Register 4
    eoc.eocReg4[0], eoc.eocReg4[1], eoc.eocReg4[2], eoc.eocReg4[3],
    eoc.eocReg4[4], eoc.eocReg4[5]);

  printf("eocReg5[32](%08x)=%02x %02x %02x %02x %02x %02x\n",    // Host/Dsp Write, vendor specific EOC Register 5
    (unsigned int) eoc_p->eocReg5,      // Host/Dsp Write, vendor specific EOC Register 4
    eoc.eocReg5[0], eoc.eocReg5[1], eoc.eocReg5[2], eoc.eocReg5[3],
    eoc.eocReg5[4], eoc.eocReg5[5]);

  printf("vendorId[8](%08x)=%02x %02x %02x %02x %02x %02x %02x %02x\n",
    (unsigned int) eoc_p->vendorId,
    eoc.vendorId[0], eoc.vendorId[1], eoc.vendorId[2], eoc.vendorId[3],
    eoc.vendorId[4], eoc.vendorId[5], eoc.vendorId[6], eoc.vendorId[7]);               // Host write

  printf("revNumber[32](%08x):%02x %02x %02x %02x\n", (unsigned int) eoc_p->revNumber,
    eoc_p->revNumber[0], eoc.revNumber[1], eoc.revNumber[2], eoc.revNumber[3]);              // Host, ATU-R Revision Number

  return 1;
}

void shutdownDsl(void)
{
  int ret;
  ret=dslhal_api_dslShutdown(ptidsl);
}


/**
 * \brief     Showing info of SHOWTIME
 * \warning   This function assume dslhal_api_gatherStatistics() has been called
 */
void display_ShowTimeInfo (tidsl_t *ptidsl)
{
  unsigned int trellis_r;
//  int offset_Trel_msgc[2] = {7, 1};
  int offset_Trel_msgr[2] = {8, 1};
  unsigned short psd_mask_qualifier = ptidsl->AppData.psd_mask_qualifier;
  unsigned short annex_selected = ptidsl->AppData.annex_selected;
  int S_ds, S_us; // UR8_MERGE_START_END CQ10470 Manjula K

  if (ptidsl->AppData.TrainedMode != T1413_ANSI) // check if GHS
  {
    printf("RT:%c%c%c%c,  CO:%c%c%c%c\n",
      isalnum(ptidsl->AppData.ghsATURVendorId[2]) ? ptidsl->AppData.ghsATURVendorId[2] : ' ',
      isalnum(ptidsl->AppData.ghsATURVendorId[3]) ? ptidsl->AppData.ghsATURVendorId[3] : ' ',
      isalnum(ptidsl->AppData.ghsATURVendorId[4]) ? ptidsl->AppData.ghsATURVendorId[4] : ' ',
      isalnum(ptidsl->AppData.ghsATURVendorId[5]) ? ptidsl->AppData.ghsATURVendorId[5] : ' ',
      isalnum(ptidsl->AppData.ghsATUCVendorId[2]) ? ptidsl->AppData.ghsATUCVendorId[2] : ' ',
      isalnum(ptidsl->AppData.ghsATUCVendorId[3]) ? ptidsl->AppData.ghsATUCVendorId[3] : ' ',
      isalnum(ptidsl->AppData.ghsATUCVendorId[4]) ? ptidsl->AppData.ghsATUCVendorId[4] : ' ',
      isalnum(ptidsl->AppData.ghsATUCVendorId[5]) ? ptidsl->AppData.ghsATUCVendorId[5] : ' ' );
  }

  // display showtime info
  printf("US Rate=%u, DS Rate=%u\n",
    ptidsl->AppData.USConRate, ptidsl->AppData.DSConRate);

//  trellis_c = dsl_generic_read(2, (unsigned int *)&offset_Trel_msgc) & 0xFF;
  trellis_r = dslhal_support_readOffset(ptidsl, 2, (unsigned int *)&offset_Trel_msgr) & 0xFF;

  // UR8_MERGE_START CQ11247_TR69_DS_LATN_SATN  YW
  if (dslhal_api_getHighPrecision())
  {
//  UR8_MERGE_START   CQ11228 HZ    
    printf("USMarg=%c%d.%d, DSMarg=%c%d.%d, FM=%d, USAtt=%u.%d, DSAtt=%u.%d, TR=%s\n",
    (((ptidsl->AppData.usMargin)/10==0)&&((ptidsl->AppData.usMargin)<0))?('-'):(' '),
    (ptidsl->AppData.usMargin)/10,   (ptidsl->AppData.usMargin>=0)?((ptidsl->AppData.usMargin)%10):((-ptidsl->AppData.usMargin)%10),
    (((ptidsl->AppData.dsMargin)/10==0)&&((ptidsl->AppData.dsMargin)<0))?('-'):(' '),
    (ptidsl->AppData.dsMargin)/10,   (ptidsl->AppData.dsMargin>=0)?((ptidsl->AppData.dsMargin)%10):((-ptidsl->AppData.dsMargin)%10),
    ptidsl->AppData.FrmMode,
    (ptidsl->AppData.usLineAttn)/10, (ptidsl->AppData.usLineAttn)%10,
    (ptidsl->AppData.dsLineAttn)/10, (ptidsl->AppData.dsLineAttn)%10,
    trellis_r ? "on" : "off");
//  UR8_MERGE_END   CQ11228 HZ    
  }
  else
  {
//  UR8_MERGE_START_END   CQ11228 HZ  Change the print format from %u to %d for the updated variables 'usMargin' and 'dsMargin'.
    printf("USMarg=%d, DSMarg=%d.%d, FM=%d, USAtt=%u.%d, DSAtt=%u.%d, TR=%s\n",
    ptidsl->AppData.usMargin,
    (ptidsl->AppData.dsMargin)>>1, (ptidsl->AppData.dsMargin&1)?5:0,
    ptidsl->AppData.FrmMode,
    (ptidsl->AppData.usLineAttn)>>1, (ptidsl->AppData.usLineAttn&1)?5:0,
    (ptidsl->AppData.dsLineAttn)>>1, (ptidsl->AppData.dsLineAttn&1)?5:0,
    trellis_r ? "on" : "off");
  }

  printf("HYB=%d, TPath=%d, ",
    ptidsl->AppData.currentHybridNum,
    ptidsl->AppData.TrainedPath);

  internal_show_agc();

  printf("PSD=0x%04x", psd_mask_qualifier);
// UR8_MERGE_START PeterHou
  if (annex_selected & (ANNEXA|ANNEXC))
// UR8_MERGE_END
    putchar('\n');
  else {
    printf(" (");
    printf("spctrm_ovlp:%c",
      (psd_mask_qualifier & OVERLAPPED_SPECTRUM_USED) ? 'Y': 'N');
    if (annex_selected & (ANNEXM |ANNEXJ))
    {
      printf(", AnxMJ_InBand_shaping_support:%c",
        (psd_mask_qualifier & ANNEXMJ_US_INBAND_SPECSHAPING_USED) ? 'Y' : 'N');
        printf(", AnxMJ:EU%d", ptidsl->AppData.max_us_tones);
    }

    if (annex_selected & ANNEXL)
        printf(", AnxL:%s", (psd_mask_qualifier & NB_ANNEXL) ? "NB" : "WB");

    if (annex_selected & ANNEXB)
      printf(", AnxB_US1to32_used:%c",
        (psd_mask_qualifier & ANNEXB_US_TONES_1TO32_USED) ? 'Y' : 'N');

    printf(")\n");
  }


  if (!(ptidsl->AppData.TrainedMode & (ADSL2_MASKS | ADSL2PLUS_MASKS)))
  { // only for g.dmt, t1413, and g.lite
    // UR8_MERGE_START CQ10470 Manjula K
    /* RSi=Ri/S */
    S_ds = ptidsl->AppData.dsSymbolsPerCodeWord ;
    S_us = ptidsl->AppData.usSymbolsPerCodeWord;
    printf("DS Rf/Ri/S/D=%d/%d/%d/%d, US Rf/Ri/S/D=%d/%d/%d/%d\n",
      ptidsl->AppData.dsFastParityBytesPerSymbol,
      S_ds ? (ptidsl->AppData.dsIntlvParityBytesPerSymbol*S_ds) :
        (ptidsl->AppData.dsIntlvParityBytesPerSymbol/2),
      ptidsl->AppData.dsSymbolsPerCodeWord,
      ptidsl->AppData.dsInterleaverDepth,
      ptidsl->AppData.usFastParityBytesPerSymbol,
     (ptidsl->AppData.usIntlvParityBytesPerSymbol*S_us),
     // UR8_MERGE_END CQ10470 
      ptidsl->AppData.usSymbolsPerCodeWord,
      ptidsl->AppData.usInterleaverDepth);
  }
  else
  { // for ADSL2\2+
    internal_showtime_display_ADSL2p(0);
  }

  if (pmc->nohost_flags>=1)
  {
    pmc->start_nohost=1;
    printf("start nohost\n");
  }

  internal_showtime_display();
}

//
// -------------- Utility Routines --------------------------------
//

//
//This function handle both decimal & hex number
//
int GetNumber (unsigned char *cp, int *number)
{
    int ccnt = 0;
    unsigned char c;
    int n = 0;
    int neg=0;

    if ((cp[0]=='0') && (toupper(cp[1])=='X'))
    {
      return GetHexNumber(&cp[2], number);
    }
    else if (*cp=='-') { // negative number
      neg=1;
      cp++;
    }

//printf("neg; neg);

    while ((c = *cp) != 0)
    {
//printf("c=0x%02x\n", c);

        n *= 10;
        if ( c >= '0' && c <= '9')
            n += c - '0';
        else
        {
            ccnt = 0;
            break;
        }
        cp++;
        ccnt++;
    }

    if (neg) n=-n;
    *number = n;
    return ccnt;
}

int GetHexNumber (unsigned char *cp, unsigned int *hex)
{
    int ccnt = 0;
    unsigned char c;
    unsigned int h = 0;

    ccnt = 0;

    while ((c = *cp) != 0)
    {
        h *= 16;

        // test and convert input to hex.
        if( (c>='0') && (c<='9') )
            h += c - '0';
        else
        if ((c>='a') && (c<='f'))
            h += c - 'a' + 10;
        else
        if ((c>='A') && (c<='F'))
            h += c - 'A' + 10;
        else
        {
            // bail out as the input is not hex
            ccnt = 0;
            break;
        }
        cp++;
        ccnt++;
    }
    *hex = h;
    return ccnt;
}

char *TrainModeFromNum (unsigned int num)
{
  int i;
  static char *strUnknownMode="UnknownMode";

  for (i=0; i<pmc->TraingModeTbl_entries; i++)
  {
    if (num==TrainModeTbl[i].id)
      return TrainModeTbl[i].str;
  }
  return strUnknownMode;  // unknow mode
}

char *AnnexFromNum (unsigned int annex_selected)
{
static char *pUnknown= "Anx:Unknown";
static struct {
  char *annex_str;
  unsigned int bitMask;
} annex_Table[] =
{
  {"AnxA", 0x1},
  {"AnxB", 0x2},
  {"AnxC", 0x4},
  {"AnxI", 0x8},
  {"AnxJ", 0x10},
  {"AnxL", 0x20},
  {"AnxM", 0x40},
};
  unsigned int i;

  for (i=0;i<NUM_ELMS(annex_Table); i++)
  {
    if (annex_Table[i].bitMask & annex_selected)
      return annex_Table[i].annex_str;
  }
  return pUnknown;
}

unsigned int byteSwap32(unsigned int in32Bits)
{
    int out32Bits;

#ifdef EB
    unsigned char tmp;
    unsigned char *pchar;
#endif

    out32Bits = in32Bits;

#ifdef EB
    pchar = (unsigned char *)(&out32Bits);

    tmp = *pchar;
    *pchar = *(pchar + 3);
    *(pchar + 3) = tmp;

    tmp = *(pchar + 1);
    *(pchar + 1) = *(pchar + 2);
    *(pchar + 2) = tmp;
#endif

    return out32Bits;

} /* end of dslhal_support_byteSwap32() */

unsigned short byteSwap16(unsigned short in16Bits)
{
    unsigned short out16Bits;

#ifdef EB
    unsigned char *pchar;
    unsigned char tmp;
#endif

    out16Bits = in16Bits;

#ifdef EB
    pchar = (unsigned char *)(&out16Bits);

    tmp = *pchar;
    *pchar = *(pchar + 1);
    *(pchar + 1) = tmp;
#endif

    return out16Bits;
}

///\endif 
