/****************************************************************************************
 *
 * FILE NAME:       tn7efm.h
 *
 * DESCRIPTION:     Linux EFM module common definition header file.
 *
 * (C) Copyright 2007, Texas Instruments Inc
 *
 * History:
 *   2/20/2007 CPH  CQ11466: Written, EFM support.
 *  10/21/2009 EYin SMS01234914: Add PTM Pre-emption support.
 *
 ****************************************************************************************/

#ifndef __TN7EFM_H__
#define __TN7EFM_H__

// #define AR7_EFM_PREEMPTION

#ifndef irqreturn_t
typedef void irqreturn_t;
#endif

#include "efmcrc.h"

// two channel: 1)normal packets 2)pre-emption(high priority)packets
#define NUM_EFM_CHAN    2 
#define MAX_EFM_CHAN    2 
#define NUM_EFM_QUEUE   2 


#define TX_EFM_QUEUE_NUM  NUM_EFM_QUEUE

// The default EFM Tx/Rx buffers, could be override by env vars
#define EFM_TX_SERVICE_MAX_DEFAULT    20
#define EFM_RX_SERVICE_MAX_DEFAULT    40
#define EFM_TX_BUFFER_NUM_DEFAULT     160
#define EFM_RX_BUFFER_NUM_DEFAULT     120
#define EFM_IPACEMAX_DEFAULT          6

#define EFM_DEFAULT_MAC "\0TEFM0"


#define AR7_EFM_DEFAULT_TIMEOUT       10    // in jiffies

#define EGRESS_TRAILOR_LEN      4
#define TC_CRC16_LEN            2

#ifdef ETH_TC_CRC16
#define EFM_TRAILOR_LEN         (EGRESS_TRAILOR_LEN+TC_CRC16_LEN) // add 2 bytes for TC_CRC16
#else
#define EFM_TRAILOR_LEN         EGRESS_TRAILOR_LEN
#endif


#define EFM_REG_OK      1
#define EFM_REG_FAILED  0

// Define EFM load. For debugging purpose,there could be more than one EFM load co-exist in the system, 
// based on efm_ctl[7:0], the selected one will be used. 
#define EFM_LOAD_NORM   1   // Normal EFM load
#ifdef EFM_DEBUG
#define EFM_LOAD_NELB   4   // Near end loopback EFM load
#endif


enum
{
  PACKET_TYPE_AAL5,
  PACKET_TYPE_NULL,
  PACKET_TYPE_OAM,
  PACKET_TYPE_TRANS,
  PACKET_TYPE_AAL2
};


// The following definition is duplicate in dsl_hal_api.h & dev_host_interface.h
#ifndef TC_MODE
#define TC_MODE
enum {
  TC_MODE_STM,
  TC_MODE_ATM,
  TC_MODE_PTM,  // EFM mode
};
#endif


// This structure contains counts will be zero out for every retrain
typedef struct _efm_stats
{
  unsigned int DevInt_count;
  unsigned int DevInt_TH_count;
  unsigned int DevInt_TL_count;
  unsigned int DevInt_Rx_count;
  unsigned int DevInt_STS_count;
  
  unsigned int send_count;
  unsigned int queue_packet_to_sar_count;
  
#ifdef EFM_DEBUG  
  unsigned int rx_crc32_dropped;
  unsigned int rx_crc32_undropped;
  unsigned int rx_crc32_dropped_len;  
#endif  
  
} EFM_STATS;

#ifdef _DGPRINTF_
#ifdef DEBUG_BUILD
#define dgprintf dprintf
#else
#define dgprintf(uDbgLevel, szFmt, args...)
#endif
#endif


void dprintf( int uDbgLevel, char * szFmt, ...);


// ----------- Function Prototype ------------------------------
int tn7efm_detect (void);
void tn7efm_initenv (void);
int tn7efm_receive (void *os_dev, int ch, unsigned int packet_size,
                    void *os_receive_info, void *data);
void tn7efm_make_crc32_table (void);
unsigned long efm_crc32 (unsigned char *buf, int len);
void tn7efm_exit (void);
int tn7efm_send_complete (void *skb1);
void tn7efm_showtime_init (void);
int tn7efm_get_currTCmode(void);

// mode switch
void tn7efm_modeSwitch_teardownATM (void);
void tn7efm_modeSwitch_teardownEFM (void);
void tn7efm_modeSwitch_initATM (void);
void tn7efm_modeSwitch_initEFM (void);
int tn7efm_modeSwitch_closeEFM (void);
int tn7efm_modeSwitch_closeATM (void);

#ifdef EFM_DEBUG
void xxdump (unsigned char* cp, int  length, char*  prefix );
#endif
// --------------------------------------------------------------

#endif // __TN7EFM_H__